from __future__ import print_function

from sys import exit
from wellpapp import Client, Config, Post, CommentWrapper, RawWrapper, raw_exts
import re
from os.path import exists, dirname
from os import makedirs, stat
from PIL import Image
from sys import stdout
from subprocess import call
from io import open

def main(arg0, argv):
	def usage():
		print("Usage:", arg0, "[-t] spec-file")
		print("spec-file contains:")
		print("source server port")
		print("destination server port")
		print("max_side")
		print("dest_dir for images")
		print("empty line")
		print("any number of search-specs")
		print("empty line")
		print("any number of include:tag and exclude:tag lines")
		exit(1)

	_guid_re = re.compile(r"-?(?:\w{6}-){3}\w{6}$")
	def parse_tag(t, **a):
		g = src_client.find_tag(t, **a)
		if g: t = g
		assert _guid_re.match(t)
		return t

	def parse_search(s):
		want = []
		dont = []
		for t in s.split():
			g = parse_tag(t, with_prefix=True)
			if g[0] == "-":
				dont.append(g[1:])
			else:
				want.append(g)
		return dict(guids=want, excl_guids=dont)

	class Spec:
		def __init__(self, fh):
			fh = CommentWrapper(fh, allow_empty=True)
			self.src = fh.readline().strip().split()
			self.dest = fh.readline().strip().split()
			self.max_side = int(fh.readline())
			self.dest_dir = fh.readline().strip()
			assert self.dest_dir != "" and exists(self.dest_dir)
			empty = fh.readline().strip()
			assert empty == ""
			line = fh.readline().strip()
			self.search = []
			while line:
				self.search.append(line)
				line = fh.readline().strip()
			self.include = []
			self.exclude = []
			fh.allow_empty = False
			for line in fh:
				ie, t = line.strip().split(":", 1)
				if ie == "include":
					self.include.append(t)
				elif ie == "exclude":
					self.exclude.append(t)
				else:
					print("bad line:", line)
					exit(1)

		def parse(self):
			self.search = list(map(parse_search, self.search))
			self.include = set(map(parse_tag, self.include))
			self.exclude = set(map(parse_tag, self.exclude))
			both = self.include.intersection(self.exclude)
			if both:
				print("Either exclude or include, not both:")
				print("\n".join(sorted([src_client.get_tag(t).name for t in both])))
				exit(1)

	def rotate_image(img, rot):
		rot = int(rot or 0)
		rotation = {90: Image.ROTATE_270, 180: Image.ROTATE_180, 270: Image.ROTATE_90}
		if rot not in rotation: return img
		return img.transpose(rotation[rot])

	def copyfile(src_fn, dest_fn):
		src_fh = open(src_fn, "rb")
		dest_fh = open(dest_fn, "wb")
		dest_fh.write(src_fh.read())
		dest_fh.close()
		src_fh.close()

	def copyexif(src_fn, dest_fn):
		call(["exiftool", "-q", "-overwrite_original", "-tagsFromFile", src_fn,
		      "-Orientation=Normal", dest_fn])

	def newer(a, b):
		return stat(a).st_mtime > stat(b).st_mtime

	if len(argv) == 1:
		test = False
		fn = argv[0]
	elif len(argv) == 2:
		test = True
		fn = argv[1]
		if argv[0] != "-t":
			usage()
	else:
		usage()

	spec = Spec(open(fn, "r", encoding="utf-8"))

	src_cfg = Config()
	src_cfg.server, src_cfg.port = spec.src
	src_client = Client(src_cfg)

	dest_cfg = Config()
	dest_cfg.server, dest_cfg.port = spec.dest
	dest_cfg.image_base = spec.dest_dir
	dest_cfg.thumb_base = spec.dest_dir
	dest_client = Client(dest_cfg)

	spec.parse()

	posts = {}
	for s in spec.search:
		ps = src_client.search_post(wanted=["tagguid", "tagname", "implied", "ext", "width", "height", "rotate", "tagdata", "datatags"], **s)
		for p in ps:
			posts[p.md5] = p
	md5s = set(posts.keys())
	posts = list(posts.values())

	datatags = set(["aaaaaa-aaaac8-faketg-bddate", "aaaaaa-aaaadt-faketg-gpspos"])
	magic_tags = set(["aaaaaa-aaaads-faketg-create", "aaaaaa-aaaaas-faketg-chaage",
	                  "aaaaaa-aaaac8-faketg-bddate", "aaaaaa-aaaaeL-faketg-bbredd",
	                  "aaaaaa-aaaaf9-faketg-heyght", "aaaaaa-aaaacr-faketg-FLekst",
	                  "aaaaaa-aaaade-faketg-rotate", "aaaaaa-aaaadt-faketg-gpspos",
	                ])
	spec.include.update(magic_tags)

	bad = set()
	for post in posts:
		for t in post.tags:
			if t.guid not in spec.include and t.guid not in spec.exclude:
				bad.add(t.name)

	spec.include.difference_update(magic_tags)

	if bad:
		print("Unspecified tags:")
		for n in bad:
			print("include:" + n)
		for n in bad:
			print("exclude:" + n)
		exit(1)

	if test: exit(0)

	z = spec.max_side

	print("Copying/rescaling images")
	count = 0
	for post in posts:
		m = post.md5
		w, h = post.width, post.height
		src_fn = src_client.image_path(m)
		dest_fn = dest_client.image_path(m)
		if not exists(dest_fn):
			d = dirname(dest_fn)
			if not exists(d): makedirs(dirname(dest_fn))
			if w > z or h > z or post.rotate > 0 or post.ext in raw_exts:
				img = Image.open(RawWrapper(open(src_fn, "rb")))
				if w > z or h > z:
					img.thumbnail((z, z), Image.ANTIALIAS)
				img = rotate_image(img, post.rotate)
				ext = post.ext
				if ext in raw_exts: ext = "jpeg"
				if ext == "jpeg":
					opts = dict(quality=90)
				else:
					opts = {}
				img.save(dest_fn, format=ext.upper(), **opts)
				if ext == "jpeg":
					copyexif(src_fn, dest_fn)
			else:
				copyfile(src_fn, dest_fn)
		src_fn = src_client.thumb_path(m, 200)
		dest_fn = dest_client.thumb_path(m, 200)
		if not exists(dest_fn) or newer(src_fn, dest_fn):
			d = dirname(dest_fn)
			if not exists(d): makedirs(dirname(dest_fn))
			copyfile(src_fn, dest_fn)
		count += 1
		progress = "\r%d/%d" % (count, len(posts))
		stdout.write(progress)
		stdout.flush()
	print()

	print("Copying to destination server..")

	if dest_client._search_post("SP", []):
		print("ERROR: destination server has posts")
		exit(1)
	dest_tags = set(t.guid for t in dest_client.find_tags("EAI", ""))
	if dest_tags.difference(magic_tags):
		print("ERROR: destination server has tags")
		exit(1)

	dest_client.begin_transaction()

	print("Tags..")
	ordered = set()
	for t in spec.include:
		tag = src_client.get_tag(t)
		dest_client.add_tag(tag.name, tag.type, tag.guid, tag.valuetype)
		if "alias" in tag:
			for alias in tag.alias:
				dest_client.add_alias(alias, tag.guid)
		if tag.ordered: ordered.add(tag.guid)

	for t in spec.include:
		for impl in src_client.tag_implies(t):
			if impl.guid in spec.include:
				dest_client.add_implies(t, *impl)

	print("Posts..")
	for post in posts:
		full = [(t.guid, t.value) for t in post.fulltags if t.guid in spec.include]
		weak = [(t.guid, t.value) for t in post.weaktags if t.guid in spec.include]
		data = Post(post)
		for t in post.datatags:
			if t.guid in datatags:
				full.append((t.guid, t.value))
			if t.guid == "aaaaaa-aaaads-faketg-create":
				data.created = t.value
		for d in ("tags", "fulltags", "weaktags", "implfulltags", "implweaktags", "settags", "impltags", "datatags"):
			del data[d]
		w, h = map(int, (data.width, data.height))
		if w > z:
			h = h * z // w
			w = z
		if h > z:
			w = w * z // h
			h = z
		data.width, data.height = w, h
		data.rotate = 0
		if data.ext in raw_exts: data.ext = "jpeg"
		dest_client.add_post(**data)
		dest_client.tag_post(post.md5, full, weak)

	rels = set()
	print("Relationships..")
	for post in posts:
		all = []
		for rel in src_client.post_rels(post.md5) or []:
			if rel + post.md5 not in rels and rel in md5s:
				rels.add(post.md5 + rel)
				all.append(rel)
		if all:
			dest_client.add_rels(post.md5, all)

	print("Ordering..")
	for g in ordered:
		order = [p.md5 for p in src_client.search_post(guids=[g], order=["group"])]
		dest_client.order(g, [m for m in order if m in md5s])

	dest_client.end_transaction()



from __future__ import print_function

from wellpapp import Client

def main(arg0, argv):
	if len(argv) != 1:
		print("Usage:", arg0, "alias")
		return 1
	client = Client()
	client.remove_alias(argv[0])



from __future__ import print_function

from os import walk, readlink, stat
from os.path import join
from wellpapp.client import Client
from wellpapp.util import identify_raw, RawWrapper, raw_exts
from io import open

def main(arg0, argv):
	if len(argv) < 1 or argv[0][0] == "-":
		print("Usage: " + arg0 + " . or raw or post-spec [post-spec [..]]")
		print("\t.         - Regenerate full cache")
		print("\traw       - Regenerate cache for all raw images")
		print("\tpost-spec - add post-spec to cache")
		print("Run in image_base")
		return 1

	def add(m, fn):
		try:
			dest = readlink(fn)
			s = stat(dest)
			z = s.st_size
			mt = int(s.st_mtime)
			fh = open(dest, "rb")
			if identify_raw(fh):
				fh.seek(0)
				jfh = RawWrapper(fh, True)
				jfh.seek(0, 2)
				jz = jfh.tell()
				jfh.close()
				l = "1 %s %d %d %d %s\n" % (m, z, mt, jz, dest)
			else:
				l = "0 %s %d %d %s\n" % (m, z, mt, dest)
			fh.close()
			res.write(l)
		except Exception:
			print(m, "failed")

	res = open("cache", "a", encoding="utf-8")
	if argv == ["."]:
		for dp, dns, fns in walk("."):
			for n in [n for n in fns if len(n) == 32]:
				add(n, join(dp, n))
	elif argv == ["raw"]:
		client = Client()
		ms = []
		for ext in raw_exts:
			p = client.search_post(guids=[("aaaaaa-aaaacr-faketg-FLekst", ext)])
			ms += [p.md5 for p in p]
		for m in ms:
			add(m, client.image_path(m))
	else:
		client = Client()
		for n in argv:
			m = client.postspec2md5(n)
			if m:
				add(m, client.image_path(m))
			else:
				print("Failed to convert " + n + " to post")
	res.close()



from base64 import b64encode, b64decode
from functools import partial
import sys

__all__ = ("_uni", "_uniw", "_enc", "_dec")

if sys.version_info[0] > 2:
	unicode = str

def _uni(s, allow_space=True):
	if not isinstance(s, unicode):
		try:
			s = s.decode("utf-8")
		except UnicodeDecodeError:
			s = s.decode("iso-8859-1")
	if not allow_space: assert u" " not in s
	return s

_uniw = partial(_uni, allow_space=False)

def _utf(s, allow_space=False):
	s = _uni(s, allow_space)
	return s.encode("utf-8")

def _enc(s):
	s = _utf(s, True)
	while len(s) % 3: s += b"\x00"
	return b64encode(s, b"_-").decode("ascii")

def _dec(enc):
	if not enc: return u""
	enc = _utf(enc)
	s = b64decode(enc, b"_-")
	while s.endswith(b"\x00"): s = s[:-1]
	return s.decode("utf-8")



from __future__ import print_function

from wellpapp import Client

def main(arg0, argv):
	if len(argv) != 2:
		print("Usage:", arg0, "into-tag from-tag")
		return 1

	client = Client()
	into_t, from_t = map(client.find_tag, argv)
	if not into_t or not from_t:
		print("Tag not found")
		return 1
	client.merge_tags(into_t, from_t)



from __future__ import print_function

from sys import exit
from wellpapp import Client, InheritValue, vtparse
from optparse import OptionParser

def main(arg0, argv):
	p = OptionParser(usage="Usage: %prog [options] set_tag implied_tag [priority]", prog=arg0)
	p.add_option("-d", "--delete", help="Delete implication", action="store_true")
	p.add_option("-p", "--priority", help="Priority of implication", type="int")
	p.add_option("-n", "--negative", help="Negative implication", action="store_true")
	p.add_option("-f", "--filter", help="Filter on set_tag value")
	p.add_option("-v", "--value", help="Value of implied tag. '' means inherit.")
	opts, args = p.parse_args(argv)

	def usage():
		p.print_help()
		exit(1)

	if len(args) not in (2, 3):
		usage()

	client = Client()
	set_tag = client.find_tag(args[0])
	implied_tag = client.find_tag(args[1])
	if len(args) == 3:
		try:
			priority = int(args[2])
			assert opts.priority is None
		except Exception:
			usage()
	else:
		priority = opts.priority or 0

	if set_tag and implied_tag:
		if opts.negative:
			implied_tag = "-" + implied_tag
		if opts.filter:
			filter = opts.filter
			if len(filter) > 1 and filter[1] in "=~":
				comp = filter[:2]
				filter = filter[2:]
			else:
				comp = filter[0]
				filter = filter[1:]
			if comp not in ("=", "<", ">", "<=", ">=", "=~"):
				print("Unknown filter comparison")
				exit(1)
			tag = client.get_tag(set_tag)
			if not tag.valuetype:
				print("Set tag does not take a value")
				exit(1)
			filter = (comp, vtparse(tag.valuetype, filter, True))
		else:
			filter = None
		if opts.value is None:
			value = None
		else:
			if opts.value:
				if opts.delete:
					print("Don't specify value when deleting")
					exit(1)
				tag = client.get_tag(implied_tag)
				if not tag.valuetype:
					print("Implied tag does not take a value")
					exit(1)
				value = vtparse(tag.valuetype, opts.value, True)
			else:
				value = InheritValue
		if opts.delete:
			client.remove_implies(set_tag, implied_tag, filter)
		else:
			client.add_implies(set_tag, implied_tag, priority, filter, value)
	else:
		print("Not found")



from __future__ import print_function

from wellpapp import Client

def main(arg0, argv):
	if len(argv) < 2:
		print("Usage:", arg0, "rotation post-spec [post-spec [..]]")
		return 1

	new_r = int(argv[0])
	if abs(new_r) not in (90, 180, 270):
		print("Can only handle 90, 180 or 270 degrees rotation")
		return 1

	def rotate(spec):
		global bad
		md5 = client.postspec2md5(spec)
		post = None
		if md5:
			post = client.get_post(md5, wanted = ["rotate", "ext", "width", "height"])
		if not post:
			print(spec + ": Post not found")
			bad = 1
			return
		r = int(post.get("rotate", 0))
		if r not in (0, 90, 180, 270):
			print(spec + ": Invalid current rotation")
			bad = 1
			return
		dest_r = (r + new_r) % 360
		props = {"rotate": dest_r}
		if abs(new_r) in (90, 270):
			props["width"], props["height"] = post["height"], post["width"]
		client.save_thumbs(md5, None, post.ext, dest_r, True)
		client.modify_post(md5, **props)

	client = Client()
	client.begin_transaction()
	bad = 0

	for spec in argv[1:]:
		rotate(spec)
	return bad



from __future__ import print_function

from wellpapp import Client
from os.path import exists, lexists
from os import unlink

def main(arg0, argv):
	if len(argv) < 1:
		print("Usage:", arg0, "[-f] post-spec [post-spec [..]]")
		print("\t-f: Force delete, even if post has tags/rels.")
		return 1

	def rmthumbs(post):
		m = post["md5"]
		sizes = list(map(int, client.cfg.thumb_sizes.split())) + ["normal", "large"]
		for z in sizes:
			if isinstance(z, int):
				path = client.thumb_path(m, z)
			else:
				path = client.pngthumb_path(m, post["ext"], z)
			if exists(path):
				unlink(path)

	def rmimg(m):
		path = client.image_path(m)
		if lexists(path):
			unlink(path)

	def delete_post(m, force):
		post = client.get_post(m, True, ["tagguid", "ext", "implied"])
		if not post:
			print(m, "post not found")
			return 1
		if post.tags:
			if force:
				client.tag_post(m, remove_tags=[t.guid for t in post.settags])
			else:
				print(m, "post has tags")
				return 1
		rels = client.post_rels(m)
		if rels:
			if force:
				client.remove_rels(m, rels)
			else:
				print(m, "post has related posts")
				return 1
		client.delete_post(m)
		rmthumbs(post)
		rmimg(m)
		return 0

	client = Client()
	ret = 0
	force = False
	for ps in argv:
		if ps == '-f':
			force = True
			continue
		m = client.postspec2md5(ps)
		if m:
			ret |= delete_post(m, force)
		else:
			print("Post not found:", ps)
			ret = 1
	return ret



from __future__ import print_function

import fuse
import stat
import errno
import os
import sys
from wellpapp import Client, Tag, raw_exts
import re
from time import time, sleep
from hashlib import md5
from struct import pack, unpack
from zlib import crc32
from xml.sax.saxutils import escape as xmlescape
from os.path import exists
from threading import Thread, RLock, Lock
from collections import namedtuple

if not hasattr(fuse, "__version__"):
	raise RuntimeError("No fuse.__version__, too old?")
fuse.fuse_python_api = (0, 2)

if sys.version_info[0] == 2:
	PY3 = False
else:
	PY3 = True
	unicode = str
	if fuse.__version__ < "1.0.0":
		raise RuntimeError("Needs at least fuse 1.0.0 on python 3")

md5re = re.compile(r"^(?:\d{6}\.)?([0-9a-f]{32})\.(\w+)$")
shortmd5re = re.compile(r"^(?:\d{6}\.)?([0-9a-f]{32})$")
metamd5re = re.compile(r"^(?:\d{6}\.)?([0-9a-f]{32})\.(\w+)\.gq\.xmp$")
sre = re.compile(r"[ /]+")
orient = {0: 1, 90: 6, 180: 3, 270: 8}
default_range = (0, 10000)

_stat_t = namedtuple("stat_t", ["version", "size", "mtime", "dest", "jpegsize"])
_search_t = namedtuple("search_t", ["want", "dontwant", "order", "range", "clean"])

def NOTFOUND():
	raise IOError(errno.ENOENT, "Not found")

class WpStat(fuse.Stat):
	def __init__(self, mode, nlink, size, time):
		self.st_mode = mode
		self.st_ino = 0
		self.st_dev = 0
		self.st_nlink = nlink
		self.st_uid = 0
		self.st_gid = 0
		self.st_size = size
		self.st_atime = time
		self.st_mtime = time
		self.st_ctime = time

def _str_tagcmp(t):
	if t[-1] is not None:
		return u"".join(t[:-1] + (t[-1].format(),))
	else:
		return t[0]

class Cache:
	def __init__(self, ttl):
		self._data = {}
		self._time = time()
		self._ttl  = ttl
		self._lock = RLock()

	def get(self, get_key, getter):
		# The VT types in the search are hashable, but compare fuzzily, which is wrong here
		key = (
			tuple(_str_tagcmp(tc) for tc in get_key.want),
			tuple(_str_tagcmp(tc) for tc in get_key.dontwant),
			get_key.order,
			get_key.range,
			get_key.clean,
		)
		with self._lock:
			if self._time < time() - self._ttl:
				self._clean()
			if key not in self._data:
				self._data[key] = (time(), getter(get_key))
			return self._data[key][1]

	def _clean(self):
		self._time = time()
		t = self._time - (self._ttl / 1.5)
		too_old = [k for k, v in self._data.items() if v[0] < t]
		for key in too_old:
			del self._data[key]

_thumbpaths = ([".thumblocal", "normal"], [".thumblocal", "large"])
_cfgpath = "/.wellpapprc"
_cloudname = ".cloud"
_rawext = dict(zip(raw_exts, ("Jpg", "jPg", "jpG", "JPg", "JPG", "JpG", "jPG", "jpg")))
assert len(_rawext) == len(raw_exts)
_rawext_r = {v: k for k, v in _rawext.items()}

class Wellpapp(fuse.Fuse):
	def __init__(self, *a, **kw):
		fuse.Fuse.__init__(self, *a, **kw)
		self._raw2jpeg = False
		self._default_search = None
		self.parser.add_option(mountopt="raw2jpeg", action="store_true", dest="_raw2jpeg",
		                       help="Present RAW files as JPEG")
		self.parser.add_option(mountopt="default_search", dest="_default_search",
		                       help="Default search (added to all searches)")

	def _cfg2file(self):
		cfg = self._client.cfg
		data = []
		for k, v in cfg.items():
			if not k.startswith("_"):
				data.append(k + "=" + v + "\n")
		res = "".join(sorted(data))
		if PY3:
			res = res.encode("utf-8")
		return res

	def _cache_read(self):
		self._cache_fh.seek(0, 1)
		for line in self._cache_fh:
			try:
				v, m, size, mtime, dest = line.rstrip("\n").split(" ", 4)
				if v == "1":
					jz, dest = dest.split(" ", 1)
				else:
					jz = 0
					assert v == "0"
				self._stat_cache[m] = _stat_t(int(v), int(size), int(mtime), dest, int(jz))
			except Exception:
				print("Bad line in cache:", line)

	def _cache_thread(self):
		while True:
			sleep(1)
			self._cache_read()

	def _prime_stat_cache(self):
		fn = self._client.cfg.image_base + "/cache"
		if not exists(fn): return
		try:
			print("Loading stat-cache..")
			self._cache_fh = open(fn, "r", encoding="utf-8") if PY3 else open(fn, "r")
			self._stat_cache = {}
			self._cache_read()
			self._use_cache = True
		except Exception as e:
			print("Failed to load cache:", e)

	# Starting threads doesn't work from __init__.
	def fsinit(self):
		if self._use_cache:
			t = Thread(target=self._cache_thread)
			t.name = "cache loader"
			t.daemon = True
			t.start()
		print("Ready")

	def _stat(self, m):
		if m not in self._stat_cache:
			print(m, "not in cache")
			p = self._client.image_path(m)
			dest = os.readlink(p)
			st = os.stat(dest)
			self._stat_cache[m] = _stat_t(0, st.st_size, st.st_mtime, dest, 0)
		return self._stat_cache[m]

	def getattr(self, path):
		spath = path.split("/")[1:]
		mode = stat.S_IFDIR | 0o555
		nlink = 2
		size = 0
		m = md5re.match(spath[-1])
		metam = metamd5re.match(spath[-1])
		time = 0
		if spath[-3:-1] in _thumbpaths:
			if not m or not m.group(2) != ".png": NOTFOUND()
			search = self._path2search("/" + " ".join(spath[:-3]))
			if not search: NOTFOUND()
			if search.order or self._raw2jpeg: # order specified or potentially unwrapped
				orgmd5 = self._resolve_thumb(search, spath[-1])
				if not orgmd5: NOTFOUND()
				mode = stat.S_IFREG | 0o444
				tfn = self._client.thumb_path(orgmd5[0], spath[-2])
				size = os.stat(tfn).st_size
				if search.order:
					# plus six digits and a period
					size += 7
			else:
				mode = stat.S_IFLNK | 0o444
			nlink = 1
		elif m:
			if self._use_cache:
				mode = stat.S_IFREG | 0o444
				version, size, time, dest, jpeg = self._stat(m.group(1))
				if self._raw2jpeg and spath[-1][-3:] in _rawext_r: # wrapped RAW
					size = jpeg
			else:
				mode = stat.S_IFLNK | 0o444
			nlink = 1
		elif metam:
			mode = stat.S_IFREG | 0o444
			size = len(self._generate_meta(metam.group(1)))
			nlink = 1
		elif path == "/" or spath[-1] in (".thumblocal", ".metadata") or \
		     spath[-2:] in _thumbpaths:
			pass
		elif path == _cfgpath:
			mode = stat.S_IFREG | 0o444
			nlink = 1
			size = len(self._cfgfile)
		elif spath[-1][:len(_cloudname)] == _cloudname:
			mode = stat.S_IFREG | 0o444
			nlink = 1
			size = len(self._generate_cloud(spath[:-1], spath[-1]))
		else:
			search = self._path2search(path)
			if not search: NOTFOUND()
			try:
				self._cache.get(search, self._search)
			except Exception:
				m = shortmd5re.match(spath[-1])
				if m:
					mode = stat.S_IFLNK | 0o444
					nlink = 1
				else:
					NOTFOUND()
		return WpStat(mode, nlink, size, time)

	def _generate_cloud(self, spath, fn):
		fn = fn[len(_cloudname):]
		count = 20
		if fn and fn[0] == ":":
			try:
				count = int(fn[1:])
			except ValueError:
				pass
			if count < 1: count = 1
		want, dontwant = self._path2search("/" + "/".join(spath))[:2]
		with self._client_lock:
			range = (0, count - 1 + len(want))
			tags = self._client.find_tags("EI", "", range=range, guids=want,
			                              excl_guids=dontwant, order="-post",
			                              flags="-datatag")
		want = [w[0][-27:] for w in want]
		names = [t.name for t in tags if t.guid not in want]
		res = "\n".join(names) + "\n"
		return res.encode("utf-8")

	def _generate_meta(self, m):
		data = u"""<?xml version="1.0" encoding="UTF-8"?><x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="XMP Core 4.1.1-Exiv2"><rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"><rdf:Description rdf:about="" xmlns:tiff="http://ns.adobe.com/tiff/1.0/" xmlns:dc="http://purl.org/dc/elements/1.1/" """
		with self._client_lock:
			post = self._client.get_post(m, wanted=["tagname", "tagdata", "rotate"])
		if "rotate" in post and post.rotate.value in orient:
			data += u"tiff:Orientation=\"" + unicode(orient[post.rotate.value]) + u"\""
		data += u"><dc:subject><rdf:Bag>"
		tags = [tag.pname + ((u"=" + unicode(tag.value)) if tag.value else u"") for tag in post.tags]
		data += u"".join([u"<rdf:li>" + xmlescape(tn) + u"</rdf:li>" \
		                 for tn in sorted(tags)])
		data += u"</rdf:Bag></dc:subject></rdf:Description></rdf:RDF></x:xmpmeta>"
		return data.encode("utf-8")

	def _resolve_thumb(self, search, thumbname):
		thumbmd5 = thumbname[:32]
		fns, pcache = self._cache.get(search, self._search)
		if not pcache:
			for fn in fns:
				m = md5re.match(fn)
				ext = m.group(2)
				ofn = m.group(1) + "." + _rawext_r.get(ext, ext)
				tmd5 = md5(fn.encode("utf-8")).hexdigest()
				pcache[tmd5] = (md5(ofn.encode("utf-8")).hexdigest(), fn)
		return pcache.get(thumbmd5)

	def readlink(self, path):
		path = path.split("/")[1:]
		m = md5re.match(path[-1])
		if m:
			if path[-3:-1] in _thumbpaths:
				return self._client.thumb_path(m.group(1), path[-2])
			else:
				return self._client.image_path(m.group(1))
		if path[-3:-1] not in _thumbpaths:
			m = shortmd5re.match(path[-1])
			if m: return self._client.image_path(m.group(1))
		NOTFOUND()

	def readdir(self, path, offset):
		list = [".", ".."]
		search = self._path2search(path)
		path = path.split("/")[1:]
		if path[-1] == ".thumblocal":
			list += ["normal", "large"]
		elif path == [""] or path[-2:] in _thumbpaths:
			pass
		elif search:
			try:
				list += self._cache.get(search, self._search)[0]
			except Exception:
				NOTFOUND()
		else:
			NOTFOUND()
		for e in list:
			yield fuse.Direntry(e)

	def _search(self, search):
		order = search.order
		range = search.range
		if not range: range = default_range
		assert None not in search.want
		assert None not in search.dontwant
		with self._client_lock:
			s = self._client.search_post(guids=search.want,
			                             excl_guids=search.dontwant,
			                             wanted=["ext"],
			                             order=order,
			                             range=range)
		r = []
		idx = 0
		prefix = ""
		for p in s:
			if order:
				prefix = "%06d." % (idx,)
				idx += 1
			if search.clean:
				r.append(prefix + p.md5)
			else:
				m = p.md5
				ext = p.ext
				if self._raw2jpeg and ext in _rawext and self._stat(m).jpegsize:
					r.append(prefix + m + "." + _rawext[ext])
				r.append(prefix + m + "." + ext)
		if not PY3:
			r = map(str, r)
		return r, {}

	def _path2search(self, path):
		if path == "/": return None
		want = set()
		dontwant = set()
		order = []
		first = None
		range = None
		clean = False
		nodefault = False
		for e in filter(None, sre.split(path[1:])):
			if e[0] == "-":
				with self._client_lock:
					e = self._client.parse_tag(e[1:], True)
				dontwant.add(e)
			elif e[:2] == "O:":
				o = e[2:]
				if o != "group":
					t = Tag()
					with self._client_lock:
						o = self._client.find_tag(o, t, True)
					assert t.valuetype
				order.append(o)
			elif e[:2] == "R:":
				range = tuple(map(int, e[2:].split(":")))
			elif e == "C:":
				clean = True
			elif e == "N:":
				nodefault = True
			else:
				with self._client_lock:
					e = self._client.parse_tag(e, True)
				want.add(e)
				if not first: first = e
		if self._default_search and not nodefault:
			def bare(tg):
				if tg[0] in "~!":
					return tg[1:]
				return tg
			allguids = {bare(t[0]) for t in want | dontwant if t}
			for n in ("want", "dontwant"):
				for t in getattr(self._default_search, n):
					if bare(t[0]) not in allguids:
						locals()[n].add(t)
		if "group" in order:
			want.remove(first)
			want = [first] + list(want)
		return _search_t(tuple(want), tuple(dontwant), tuple(order), range, clean)

	def main(self, *a, **kw):
		self._cache = Cache(30)
		self._client = Client()
		self._client_lock = RLock()
		self._cfgfile = self._cfg2file()
		self._use_cache = False
		self._prime_stat_cache()
		if self._raw2jpeg and not self._use_cache:
			raise Exception("raw2jpeg only works with a stat-cache")
		if self._raw2jpeg:
			from wellpapp import RawWrapper
		if self._default_search:
			ds = "/" + self._default_search
			self._default_search = None
			self._default_search = self._path2search(ds)
			if None in self._default_search.want or None in self._default_search.dontwant:
				raise Exception("Default search broken (%r)" % (self._default_search,))
		wp = self
		class FakeFile:
			keep_cache = False
			direct_io = False
			_fh = None
			data = ""
			def __init__(self, path, flags, *mode):
				rwflags = flags & (os.O_RDONLY | os.O_WRONLY | os.O_RDWR)
				if rwflags != os.O_RDONLY: NOTFOUND()
				if path == _cfgpath:
					self.data = wp._cfgfile
					return
				spath = path.split("/")
				metam = metamd5re.match(spath[-1])
				if metam:
					self.data = wp._generate_meta(metam.group(1))
					return
				if spath[-1][:len(_cloudname)] == _cloudname:
					self.data = wp._generate_cloud(spath[1:-1], spath[-1])
					return
				if spath[-3:-1] in _thumbpaths:
					self.data = self._make_thumb(spath)
				else:
					fn = spath[-1].split(".")
					m = fn[-2][-32:]
					fh = self._open(m)
					if fh:
						if wp._raw2jpeg and fn[-1] in _rawext_r:
							# @@ Check wp._stat(m).version
							self._fh = RawWrapper(fh, True)
						else:
							self._fh = fh
					self._lock = Lock()
			def _open(self, m):
				try:
					dest = wp._stat(m).dest
					return open(dest, "rb")
				except Exception:
					pass
				try:
					p = wp._client.image_path(m)
					dest = os.readlink(p)
					fh = open(dest, "rb")
					wp._stat_cache[m] = wp._stat(m)._replace(dest=dest)
					return fh
				except Exception:
					pass
			# FUSE doesn't seem to like destroying these objects.
			# But it does call release, so I'll do what I can.
			def __del__(self):
				self.release(0)
			def release(self, flags):
				if self._fh: self._fh.close()
				self.data = self._fh = self._Lock = None
			def _make_thumb(self, spath):
				search = wp._path2search("/".join(spath[:-3]))
				if not search: NOTFOUND()
				orgmd5, fn = wp._resolve_thumb(search, spath[-1])
				ext = fn.split(".")[-1]
				tfn = wp._client.thumb_path(orgmd5, spath[-2])
				fh = open(tfn, "rb")
				data = fh.read()
				fh.close()
				if not (search.order or ext in _rawext_r):
					return data
				data = data.split(b"tEXtThumb::URI\0")
				if len(data) != 2: NOTFOUND()
				pre, post = data
				clen, = unpack(">I", pre[-4:])
				if search.order: # It's longer only of search was ordered
					pre = pre[:-4] + pack(">I", clen + 7)
				post = post[clen - 7:]
				tEXt = b"tEXtThumb::URI\0" + fn.encode("utf-8")
				crc = crc32(tEXt)
				if crc < 0: crc += 0x100000000
				tEXt += pack(">I", crc)
				return pre + tEXt + post
			def read(self, length, offset):
				if self._fh:
					with self._lock:
						self._fh.seek(offset)
						return self._fh.read(length)
				else:
					return self.data[offset:offset + length]
		self.file_class = FakeFile
		return fuse.Fuse.main(self, *a, **kw)

def main(arg0, argv):
	server = Wellpapp(prog=arg0)
	server.parse(argv, errex=1, values=server)
	server.main()



from __future__ import print_function

from wellpapp import Client
from optparse import OptionParser

def main(arg0, argv):
	p = OptionParser(usage="Usage: %prog [options] tagname [tagtype]", prog=arg0)
	p.add_option("-v", "--valuetype", help="Valuetype of tag")
	opts, args = p.parse_args(argv)

	if len(args) not in (1, 2):
		p.print_help()
		return 1

	client = Client()
	client.add_tag(*args, **opts.__dict__)



import re
from fractions import Fraction
from decimal import Decimal
from abc import ABCMeta, abstractmethod, abstractproperty
from time import localtime, struct_time, strftime, gmtime
from calendar import timegm
from math import log, log10, sin, cos, acos, radians, pi
import sys

from wellpapp._util import _uni, _enc, _dec

__all__ = ("ValueType", 'VTstring', 'VTword', 'VTnumber', 'VTint', 'VTuint',
           'VTfloat', 'VTf_stop', 'VTstop', 'VTdatetime', 'VTgps', 'VTnull',
           'valuetypes',)

if sys.version_info[0] > 2:
	basestring = (bytes, str)
	long = int

class ValueType(object):
	"""Represents the value of a tag.
	v.value is the value as an apropriate type (str, float, int).
	v.exact is an exact representation of the value (str, int, Fraction).
	v.fuzz is how inexact the value is.
	v.exact_fuzz is like .exact but for the fuzz.
	v.str (or str(v)) is a string representation of exact value+-fuzz.
	v.format() is this value formated for sending to server.

	Comparisons:
	== and != compare that both value and fuzz match,
	other comparisons apply the fuzz.
	Use value.overlap(other) to check for equality with fuzz."""

	__metaclass__ = ABCMeta

	@abstractmethod
	def __init__(self): pass

	@abstractproperty
	def type(self): pass

	@abstractproperty
	def _cmp_t(self): pass

	_repr_extra = ""
	_repr = None

	str = ""
	value = 0
	exact = 0
	fuzz = None
	exact_fuzz = None

	def __setattr__(self, name, value):
		raise AttributeError("ValueTypes are immutable")
	def __delattr__(self, name):
		raise AttributeError("ValueTypes are immutable")
	def __str__(self):
		return self.str
	def __repr__(self):
		c = self.__class__
		rs = repr(self._repr) if self._repr else repr(self.str)
		return c.__module__ + "." + c.__name__ + "(" + rs + self._repr_extra + ")"
	def __hash__(self):
		return hash(self.exact) ^ hash(self.exact_fuzz) ^ hash(self.type)
	def __cmp(self, other):
		if isinstance(self, VTnumber) and isinstance(other, (int, long, float)):
			return VTnumber(other)
		if not isinstance(other, ValueType) or self._cmp_t != other._cmp_t:
			raise TypeError("Can only compare to a " + self._cmp_t)
		return other
	def __eq__(self, other):
		return type(self) == type(other) and \
		       self.exact == other.exact and \
		       self.exact_fuzz == other.exact_fuzz
	def __ne__(self, other):
		if not isinstance(other, ValueType): return True
		return type(self) != type(other) or \
		       self.exact != other.exact or \
		       self.exact_fuzz != other.exact_fuzz
	def __lt__(self, other):
		other = self.__cmp(other)
		return self.min < other.max
	def __le__(self, other):
		other = self.__cmp(other)
		return self.min <= other.max
	def __gt__(self, other):
		other = self.__cmp(other)
		return self.max > other.min
	def __ge__(self, other):
		other = self.__cmp(other)
		return self.max >= other.min

	@property
	def min(self):
		f = self.fuzz or 0
		if f < 0:
			return self.value + f
		else:
			return self.value

	@property
	def max(self):
		if self.fuzz:
			return self.value + abs(self.fuzz)
		else:
			return self.value

	def overlap(self, other):
		other = self.__cmp(other)
		return self.min <= other.max and other.min <= self.max

	def format(self):
		return self.str

class VTnull(ValueType):
	type = "null"
	_cmp_t = "VTstring"
	str = ""
	value = ""
	exact = ""

	def __init__(self, val=None, human=False):
		assert val is None or (isinstance(val, basestring) and not val)

	def __nonzero__(self):
		return False

class VTstring(ValueType):
	"""Represents the value of a tag with valuetype string.
	v.value, v.exact and v.str are all the same string.
	There is no fuzz for strings."""

	type = "string"
	_cmp_t = "VTstring"
	_repr_extra = ", True"

	def __init__(self, val, human=False):
		if human:
			val = _uni(val)
		else:
			val = _dec(val)
		for name in ("str", "value", "exact"):
			self.__dict__[name] = val
	def __unicode__(self):
		return self.str
	if sys.version_info[0] > 2:
		__str__ = __unicode__
	else:
		def __str__(self):
			return self.str.encode("utf-8")
	def format(self):
		return _enc(self.str)

class VTword(VTstring):
	"""Represents the value of a tag with valuetype word.
	v.value, v.exact and v.str are all the same string.
	There is no fuzz for words."""

	type = "word"
	_repr_extra = ""

	def __init__(self, val, human=False):
		if " " in val: raise ValueError(val)
		val = _uni(val)
		for name in ("str", "value", "exact"):
			self.__dict__[name] = val
	def format(self):
		return self.str

class VTnumber(ValueType):
	_cmp_t = "VTnumber or simple number"
	type = "simple number wrapper"

	def __init__(self, number):
		if not isinstance(number, (int, long, float)):
			raise ValueError(number)
		self.__dict__["value"] = self.__dict__["exact"] = number
		self.__dict__["fuzz"] = self.__dict__["exact_fuzz"] = 0

	def _parse(self, v, vp, vp2, fp):
		v = str(v)
		self.__dict__["str"] = v
		a = v.split("+", 1)
		self.__dict__["exact"] = vp(a[0])
		self.__dict__["value"] = vp2(self.exact)
		if len(a) == 2:
			self.__dict__["exact_fuzz"] = fp(a[1])
			self.__dict__["fuzz"] = vp2(self.exact_fuzz)
		else:
			self.__dict__["fuzz"] = self.__dict__["exact_fuzz"] = 0

	def __int__(self):
		return int(self.exact)
	def __long__(self):
		return long(self.exact)
	def __float__(self):
		return float(self.exact)

class VTint(VTnumber):
	__doc__ = ValueType.__doc__
	type = "int"

	def __init__(self, val, human=False):
		self._parse(val, int, int, int)

class VTuint(VTnumber):
	__doc__ = ValueType.__doc__
	type = "uint"

	def __init__(self, val, human=False):
		p = int if human else lambda x: int(x, 16)
		self._parse(val, p, int, int)
		if self.fuzz:
			s = "%d+%d" % (self.value, self.fuzz)
			r = "%x+%d" % (self.value, self.fuzz)
		else:
			s = str(self.value)
			r = "%x" % (self.value,)
		self.__dict__["str"] = s
		self.__dict__["_repr"] = r
	def format(self):
		return self._repr

class VTfloat(VTnumber):
	__doc__ = ValueType.__doc__
	type = "float"

	def __init__(self, val, human=False):
		def intfrac(v):
			try:
				return int(v)
			except ValueError:
				return Fraction(v)
		self._parse(val, intfrac, float, intfrac)

	def _ffix(self, value, fuzzyfuzz):
		if self.fuzz > 0:
			value -= fuzzyfuzz
			self.__dict__["fuzz"] += 2 * fuzzyfuzz
		else:
			self.__dict__["fuzz"] -= fuzzyfuzz
		self.__dict__["value"] = value

class VTf_stop(VTfloat):
	__doc__ = ValueType.__doc__
	type = "f-stop"

	def __init__(self, val, human=False):
		VTfloat.__init__(self, val, human)
		self._ffix(2.0 * log(self.value, 2.0), 0.07)

class VTstop(VTfloat):
	__doc__ = ValueType.__doc__
	type = "stop"

	def __init__(self, val, human=False):
		VTfloat.__init__(self, val, human)
		self._ffix(10.0 * log10(self.value) / 3.0, 0.01)

class VTdatetime(ValueType):
	type = "datetime"
	_cmp_t = "VTdatetime"
	_ranges = (9999, 12, 31, 23, 59, 59)
	fres = r"(\+-?(?:\d+(?:\.\d+|/\d+)?))?"
	_re = re.compile(r"(\d{4})" + fres + (r"(?:-(\d\d)" + fres) * 2 + r"(?:([T ])" + \
	#                   YYYY                   mm + dd                      T
	                 r"(\d\d)" + fres + r"(?::(\d\d)" + fres + r"(?::(\d\d)" + \
	#                    HH                     MM                     SS
	                 r")?" * 5 + r"(\+-?(?:\d+(?:\.\d+|/\d+)?[YmdHMS]?))?$")
	#                                fuzz with unit
	del fres

	def __init__(self, val, human=False):
		if isinstance(val, tuple):
			try:
				assert len(val) == 2
				int(val[0])
				if val[1]:
					int(val[1])
			except Exception:
				raise ValueError(val)
			strval = strftime("%Y-%m-%dT%H:%M:%S", gmtime(int(val[0])))
			if val[1]:
				strval = "%s+%d" % (strval, int(val[1]))
			strval += "Z"
		else:
			try:
				strval = str(val)
			except Exception:
				raise ValueError(val)
		if not strval: raise ValueError(val)
		if strval[-1] == "Z":
			zone = "Z"
			offset = 0
			strval = strval[:-1]
		else:
			zone = strval[-5:]
			try:
				assert zone[0] in "+-"
				assert False not in [c in "0123456789" for c in zone[1:]]
				hour, minute = int(zone[1:3]), int(zone[3:])
				assert 0 <= hour <= 12 and 0 <= minute < 60
				offset = (hour * 60 + minute) * 60
				if zone[0] == "+": offset = -offset
				strval = strval[:-5]
			except Exception:
				zone = None
		m = self._re.match(strval)
		if not m: raise ValueError(val)
		allval = m.groups()
		if allval[6] == " " and not human: raise ValueError(val)
		values = allval[:6] + allval[7:12]
		datev = values[::2]
		date = [int(v) if v else 1 for v in datev[:3]] + [int(v) if v else 0 for v in datev[3:]]
		if 0 in date[:3]: raise ValueError(val)
		for iv, mv in zip(date, self._ranges):
			if iv > mv: raise ValueError(val)
		fuzz = list(values[1::2])
		steps = []
		with_steps = False
		for pos, v in enumerate(fuzz):
			if v:
				v = int(v[1:])
				with_steps = True
			steps.append(v)
		parsed = struct_time(date + [0, 0, 0])
		unit = None
		last_fuzz = allval[12]
		if last_fuzz:
			if last_fuzz == "+": raise ValueError(val)
			units = ["Y", "m", "d", "H", "M", "S"]
			if last_fuzz[1] == "-":
				mult = -1
				f = last_fuzz[2:]
			else:
				mult = 1
				f = last_fuzz[1:]
			if f[-1] in units:
				unit = f[-1]
				mults = [12.0, 30.5, 24, 60, 60]
				while unit != units.pop(): mult *= mults.pop()
				f = f[:-1] or "1"
			exact_fuzz = VTfloat(f).exact * mult
		else:
			exact_fuzz = 0
		if not zone:
			lparsed = localtime(timegm(parsed))
			offset = timegm(parsed) - timegm(lparsed)
		value = timegm(parsed)
		if None in datev:
			valid_steps = datev.index(None)
			t2 = self._step_end(parsed, valid_steps)
			implfuzz = t2 - value
		else:
			implfuzz = 0
			valid_steps = 6
		fuzz = int(exact_fuzz)
		if fuzz != exact_fuzz:
			fuzz = float(exact_fuzz)
		if not unit and implfuzz: fuzz *= implfuzz * 2;
		fuzz += implfuzz
		if implfuzz and not exact_fuzz:
			exact_fuzz = implfuzz
		self._init(parsed, value, offset, exact_fuzz, fuzz, last_fuzz or "", valid_steps, zone, with_steps, steps)
		self.__dict__["str"] = strval.replace(" ", "T") + (zone or "")

	def _init(self, parsed, value, offset, exact_fuzz, fuzz, _fuzz, valid_steps, zone, with_steps, steps):
		lt = localtime(value + offset)
		lts = "%04d-%02d-%02dT%02d:%02d:%02d" % lt[:6]
		lts = lts[:valid_steps * 3 + 1]
		value += offset
		self.__dict__["_lts"] = lts
		self.__dict__["_fuzz"] = _fuzz
		self.__dict__["time"] = parsed
		self.__dict__["zone"] = zone
		self.__dict__["_tz"] = offset if zone else None
		self.__dict__["value"] = value
		self.__dict__["exact"] = value
		self.__dict__["fuzz"] = fuzz
		self.__dict__["exact_fuzz"] = exact_fuzz
		self.__dict__["utcoffset"] = offset
		self.__dict__["steps"] = tuple(steps)
		self.__dict__["with_steps"] = with_steps
		self.__dict__["valid_steps"] = valid_steps

	def localtimestr(self, include_fuzz=True):
		if not include_fuzz: return self._lts
		return self._lts + self._fuzz

	@property
	def min(self):
		if self.with_steps:
			min_l = [v + min(s or 0, 0) for v, s in zip(self.time, self.steps + (0,))]
			parsed = struct_time(min_l + [0, 0, 0])
			value = timegm(parsed) + self.utcoffset
		else:
			value = self.value
		return value + min(self.fuzz or 0, 0)

	@property
	def max(self):
		if self.with_steps:
			max_l = [v + abs(s or 0) for v, s in zip(self.time, self.steps + (0,))]
			parsed = struct_time(max_l + [0, 0, 0])
			value = timegm(parsed) + self.utcoffset
		else:
			value = self.value
		return value + abs(self.fuzz or 0)

	def overlap(self, other):
		if not isinstance(other, VTdatetime):
			raise TypeError("Can only compare to a VTdatetime")
		if self.min > other.max or other.min > self.max:
			return False
		return self._cmp_step(self, other)

	def _cmp_step(self, a, b):
		if a.with_steps:
			if self.fuzz < 0:
				nsf = self.fuzz
				psf = nsf * -2
			else:
				psf = self.fuzz
				nsf = 0
			for f in range(5):
				fuzz = a.steps[f] or 0
				for i in range(min(fuzz, 0), abs(fuzz) + 1):
					l = list(self.time)
					l[f] += i
					unixtime = timegm(l)
					fuzz = psf
					if self.valid_steps < 6:
						t2 = self._step_end(l, self.valid_steps)
						fuzz += t2 - unixtime
					value = unixtime + nsf + self.utcoffset
					a_min = value + min(fuzz, 0)
					a_max = value + abs(fuzz)
					if a is self:
						if self._cmp_step(b, (a_min, a_max)):
							return True
					else:
						if a_min <= b[1] and b[0] <= a_max:
							return True
		else:
			if a is self:
				return self._cmp_step(b, (a.min, a.max))
			else:
				return a.min <= b[1] and b[0] <= a.max
		return False

	def _step_end(self, l, valid_steps):
		l = list(l)
		for i in range(valid_steps, len(self._ranges)):
			l[i] = self._ranges[i]
		if valid_steps == 2: # year + month specified
			# "day 0" of next month
			l[1] += 1
			l[2] = 0
			if l[1] == 13:
				l[1] = 1
				l[0] += 1
		return timegm(l)

	def __add__(self, other):
		if hasattr(other, "total_seconds"):
			other = other.total_seconds()
		value = int(round(self.exact - self.utcoffset + other))
		parsed = gmtime(value)
		res = VTdatetime((0, 0))
		res._init(parsed, value, self.utcoffset, self.exact_fuzz, self.fuzz, self._fuzz, self.valid_steps, self.zone, self.with_steps, self.steps)
		ts = "%04d-%02d-%02dT%02d:%02d:%02d" % parsed[:6]
		ts = ts[:self.valid_steps * 3 + 1]
		res.__dict__["str"] = ts + (self.zone or "")
		return res

	def __sub__(self, other):
		return self.__add__(-other)

class VTgps(ValueType):
	"""Represents the value of a tag with valuetype gps."""

	type = "gps"
	_cmp_t = "VTgps"
	altitude = None
	_R = 6371000.0 # Earth's radius in meters (approx)

	def __init__(self, val, human=False):
		try:
			strval = str(val)
			s = strval.split(",")
			if len(s) == 3:
				fields = ("lat", "lon", "altitude")
			else:
				assert len(s) == 2
				fields = ("lat", "lon")
		except (UnicodeEncodeError, AssertionError):
			raise ValueError(val)
		fc = s[-1].split("+-")
		if len(fc) > 1:
			if len(fc) != 2:
				raise ValueError(val)
			s[-1] = fc[0]
			self.__dict__["fuzz"] = float(fc[1])
			self.__dict__["exact_fuzz"] = Decimal(fc[1])
		self.__dict__["_fuzz_R"] = (self.fuzz or 0) / self._R
		for n, c in zip(fields, s):
			try:
				self.__dict__[n] = Decimal(c)
			except ValueError:
				raise ValueError(val)
		self.__dict__["exact"] = (self.lat, self.lon, self.altitude)
		for n in ("lat", "lon", "altitude"):
			v = self.__dict__.get(n)
			if v is not None:
				self.__dict__[n] = float(v)
		self.__dict__["value"] = (self.lat, self.lon, self.altitude)
		self.__dict__["str"] = strval

	def __str__(self):
		return self.str
	def format(self):
		return self.str

	@property
	def _sphere_fuzz_lon(self):
		if not self.fuzz:
			return 0
		lat = radians(self.lat)
		clat = cos(lat)
		if clat == 0:
			return pi
		slat = sin(lat)
		return acos((cos(self._fuzz_R) - slat * slat) / (clat * clat))

	@property
	def min(self):
		if not self.fuzz:
			return self.value[:2]
		return self.lat - self._fuzz_R, self.lon - self._sphere_fuzz_lon

	@property
	def max(self):
		if not self.fuzz:
			return self.value[:2]
		return self.lat + self._fuzz_R, self.lon + self._sphere_fuzz_lon

	def distance(self, other):
		""""""
		if not isinstance(other, VTgps):
			raise TypeError("Can only compare to a VTgps")
		return self._sphere_dist(other) * self._R

	def _sphere_dist(self, other):
		alat = radians(self.lat)
		blat = radians(other.lat)
		try:
			return acos(sin(alat) * sin(blat) + cos(alat) * cos(blat) * cos(radians(other.lon - self.lon)))
		except ValueError:
			return pi

	def overlap(self, other):
		if not isinstance(other, VTgps):
			raise TypeError("Can only compare to a VTgps")
		dist = self._sphere_dist(other)
		return dist <= ((self._fuzz_R or 0) + (other._fuzz_R or 0))

valuetypes = {"string"  : VTstring,
              "word"    : VTword,
              "int"     : VTint,
              "uint"    : VTuint,
              "float"   : VTfloat,
              "f-stop"  : VTf_stop,
              "stop"    : VTstop,
              "datetime": VTdatetime,
              "gps"     : VTgps,
             }



import socket
import os
import hashlib
import re
from functools import partial
from collections import namedtuple
import sys

from wellpapp.vt import VTdatetime, VTuint, VTint, VTnull, valuetypes
from wellpapp._util import _uniw, _uni
from wellpapp.util import DotDict, CommentWrapper, make_pdirs, RawWrapper

__all__ = ("Client", "Config", "Post", "Tag", "WellpappError", "ResponseError",
           "DuplicateError", "InheritValue", "ImplicationTuple", "vtparse",
          )

if sys.version_info[0] > 2:
	basestring = (bytes, str)
	def itervalues(d):
		return iter(d.values())
else:
	def itervalues(d):
		return d.itervalues()

class WellpappError(Exception): pass
class ResponseError(WellpappError): pass
class DuplicateError(WellpappError): pass

class InheritValue:
	def __str__(self):
		return "<InheritValue>"
InheritValue = InheritValue()

ImplicationTuple = namedtuple("ImplicationTuple", "guid prio filter value")

def _rfindany(s, chars, pos=-1):
	if pos < 0: pos = len(s)
	for i in range(pos - 1, -1, -1):
		if s[i] in chars: return i
	return -1

def _tagspec(type, value):
	if value[0] in u"~!":
		type = value[0] + type
		value = value[1:]
	return type + value

class Post(DotDict): pass

class Tag(DotDict):
	def _populate(self, res, flags=""):
		res = res.split()
		alias = []
		flaglist = []
		hexint = partial(int, base=16)
		dummy = lambda s: s
		incl = {u"N": ("name", dummy),
		        u"T": ("type", dummy),
		        u"A": ("alias", alias.append),
		        u"P": ("posts", hexint),
		        u"W": ("weak_posts", hexint),
		        u"F": ("flags", flaglist.append),
		        u"G": ("guid", str),
		        u"V": ("valuetype", str),
		       }
		for data in res:
			if data[0] in incl:
				name, parser = incl[data[0]]
				self[name] = parser(data[1:])
		self.alias = alias
		if flaglist:
			del self.flags
			for flag in flaglist:
				self[flag] = True
		vt = (self.valuetype or "").split("=", 1)
		if len(vt) == 2:
			self.valuetype = vt[0]
			self.value = vtparse(*vt)
		if u"~" in flags:
			if self.name: self.pname = u"~" + self.name
			if self.guid: self.pguid = "~" + self.guid
		else:
			if self.name: self.pname = self.name
			if self.guid: self.pguid = self.guid

_nonetag = Tag()

class TagDict(dict):
	"""Dictionary-like object that holds the tag lists in Posts.
	Members are keyed on both .name and .guid (if they exist).
	A .guid always shadows a .name.
	Iterates over Tag objects, not names and/or guids.
	Use .names or .guids if you want dicts with just those keys.
	(Those are normal dicts, iterating over keys.)
	"""

	def __init__(self):
		self.names = {}
		self.guids = {}

	def _add(self, tag, name, guid):
		if name:
			if name not in self:
				dict.__setitem__(self, name, tag)
			self.names[name] = tag
		if guid:
			dict.__setitem__(self, guid, tag)
			self.guids[guid] = tag

	def __len__(self):
		return max(len(self.names), len(self.guids))

	def __iter__(self):
		if self.guids: return itervalues(self.guids)
		return itervalues(self.names)

	itervalues = __iter__

	def values(self):
		if self.guids: return self.guids.values()
		return self.names.values()

	def __setitem__(self, key, value):
		raise AttributeError("TagDicts are immutable-ish")

	def __delitem__(self, key):
		raise AttributeError("TagDicts are immutable-ish")

	def get(self, key, default=_nonetag):
		"""Just like get on dict, except the default is an empty Tag.
		This is still False as a bool, but you can do things like
		post.tags.get("foo").value which gives None if foo isn't set.
		"""
		return dict.get(self, key, default)

def vtparse(vtype, val, human=False):
	try:
		return valuetypes[vtype](val, human)
	except ValueError:
		if human and val == "":
			return VTnull()
		raise

_field_sparser = {
	"created"        : VTdatetime,
	"imgdate"        : VTdatetime,
	"width"          : VTuint,
	"height"         : VTuint,
	"rotate"         : VTint,
}

_p_int = lambda i: str(int(i))
_p_hexint = lambda i: "%x" % (int(i),)
def _p_date(val):
	if isinstance(val, basestring): return _uniw(val)
	return val.format()
_field_cparser = {
	"width"          : _p_hexint,
	"height"         : _p_hexint,
	"rotate"         : _p_int,
	"created"        : _p_date,
	"imgdate"        : _p_date,
	"ext"            : _uniw,
	"MD5"            : _uniw,
}

class Config(DotDict):
	def __init__(self, local_rc=False, **kw):
		DotDict.__init__(self, tagwindow_width=840, tagwindow_height=600, **kw)
		rcs = []
		rc_name = ".wellpapprc"
		path = "/"
		if local_rc:
			rcs = [rc_name]
		else:
			rcs = [os.path.join(os.environ["HOME"], rc_name)]
			for dir in os.getcwd().split(os.path.sep):
				path = os.path.join(path, dir)
				rc = os.path.join(path, rc_name)
				if os.path.exists(rc): rcs.append(rc)
		for rc in rcs:
			self._load(rc)

	def _load(self, fn):
		with CommentWrapper(open(fn)) as fh:
			for line in fh:
				line = line.strip()
				a = line.split("=", 1)
				assert(len(a) == 2)
				self[a[0]] = a[1]

class Client:
	_prot_max_len = 4096

	def __init__(self, cfg=None):
		if not cfg:
			cfg = Config()
		self.cfg = cfg
		if cfg.socket:
			self.server = cfg.socket
		else:
			self.server = (cfg.server, int(cfg.port))
		self.is_connected = False
		self._md5re = re.compile(r"^[0-9a-f]{32}$", re.I)
		base = cfg.image_base
		while base.endswith("/"): base = base[:-1]
		base = re.escape(base)
		self._destmd5re = re.compile(r"^" + base + r"/[0-9a-f]/[0-9a-f]{2}/([0-9a-f]{32})$")

	def _reconnect(self):
		if self.is_connected: return
		if isinstance(self.server, tuple):
			self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		else:
			self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		self._sock.connect(self.server)
		self._fh = self._sock.makefile('rb')
		self.is_connected = True

	def _writeline(self, line, retry=True):
		assert u"\n" not in line
		self._reconnect()
		line = line.encode("utf-8") + b"\n"
		try:
			self._sock.sendall(line)
		except IOError:
			self.is_connected = False
			if retry:
				self._reconnect()
				self._sock.sendall(line)

	def _readline(self):
		return self._fh.readline()[:-1].decode("utf-8")

	def _parse_search(self, line, posts, wanted, props):
		if line == u"OK": return True
		if line[0] != u"R": raise ResponseError(line)
		if line[1] == u"E": raise ResponseError(line)
		if line[1] == u"R":
			if props != None: props["result_count"] = int(line[2:], 16)
			return
		f = Post()
		seen = set()
		pieces = line[1:].split(u" :")
		for token in pieces[0].split():
			type = token[0]
			data = token[1:]
			if type == u"P":
				f.md5 = str(data)
			elif type == u"F":
				field, value = data.split(u"=", 1)
				field = str(field)
				if field in _field_sparser:
					f[field] = _field_sparser[field](value)
				else:
					f[field] = value
			else:
				raise ResponseError(line)
		f.fulltags = TagDict()
		f.weaktags = TagDict()
		f.implfulltags = TagDict()
		f.implweaktags = TagDict()
		f.datatags = TagDict()
		for piece in pieces[1:-1]:
			flags, data = piece.split(u" ", 1)
			if u"I" in flags and u"~" in flags:
				ta = f.implweaktags
			elif u"I" in flags:
				ta = f.implfulltags
			elif u"~" in flags:
				ta = f.weaktags
			elif u"D" in flags:
				ta = f.datatags
			else:
				ta = f.fulltags
			t = Tag()
			t._populate(data, flags)
			ta._add(t, t.name, t.guid)
		if not f.md5: raise ResponseError(line)
		if f.md5 in seen: raise DuplicateError(f.md5)
		seen.add(f.md5)
		tagdicts = [f.fulltags, f.weaktags]
		if wanted and "implied" in wanted:
			settags = TagDict()
			for td in tagdicts:
				for tag in td:
					settags._add(tag, tag.pname, tag.pguid)
			idicts = [f.implfulltags, f.implweaktags]
			tagdicts += idicts
			impltags = TagDict()
			for td in idicts:
				for tag in td:
					impltags._add(tag, tag.pname, tag.pguid)
			f.settags = settags
			f.impltags = impltags
		else:
			del f.implfulltags
			del f.implweaktags
		if not wanted or "datatags" not in wanted:
			del f.datatags
		else:
			tagdicts.append(f.datatags)
		alltags = TagDict()
		for td in tagdicts:
			for tag in td:
				alltags._add(tag, tag.pname, tag.pguid)
		f.tags = alltags
		posts.append(f)

	def _search_post(self, search, wanted=None, props=None):
		self._writeline(search)
		posts = []
		while not self._parse_search(self._readline(), posts, wanted, props): pass
		return posts

	def get_post(self, md5, separate_implied=False, wanted=None):
		if wanted is None:
			wanted = ["tagname", "tagguid", "tagdata", "datatags", "ext", "created", "width", "height"]
		if separate_implied and "implied" not in wanted:
			wanted = ["implied"] + wanted
		search = u"SPM" + _uniw(md5) + " F".join([""] + self._filter_wanted(wanted))
		posts = self._search_post(search, wanted)
		if not posts: return None
		assert posts[0].md5 == md5
		return posts[0]

	def delete_post(self, md5):
		assert " " not in md5
		cmd = u"DP" + _uniw(md5)
		self._writeline(cmd)
		res = self._readline()
		if res != u"OK": raise ResponseError(res)

	def _list(self, data, converter=_uniw):
		if not data: return []
		if isinstance(data, basestring): return [converter(data)]
		return list(map(converter, data))

	def _guids2posneg(self, guids):
		pos, neg = [], []
		for g in guids:
			if g[0] == u"!":
				pos.append(g)
			elif g[0] == u"-":
				neg.append(g[1:])
			else:
				pos.append(g)
		return pos, neg

	def _filter_wanted(self, wanted):
		return [w for w in self._list(wanted, str) if w != "datatags"]

	def _build_search(self, guids=None, excl_guids=None, wanted=None, order=None, range=None):
		guids = self._list(guids, self._tag2spec)
		excl_guids = self._list(excl_guids, self._tag2spec)
		pos1, neg1 = self._guids2posneg(guids)
		neg2, pos2 = self._guids2posneg(excl_guids)
		search = []
		for want in self._filter_wanted(wanted):
			search += [u"F", want, u" "]
		for guid in pos1 + pos2:
			search += [u"T", _tagspec(u"G", guid), u" "]
		for guid in neg1 + neg2:
			search += [u"t", _tagspec(u"G", guid), u" "]
		for o in self._list(order, str):
			search += [u"O", o, u" "]
		if range != None:
			search += [u"R%x:%x" % tuple(range)]
		return u"".join(search)

	def search_post(self, wanted=None, props=None, **kw):
		search = u"SP" + self._build_search(wanted=wanted, **kw)
		posts = self._search_post(search, wanted, props)
		return posts

	def _fieldspec(self, **kwargs):
		f = [_uniw(f) + u"=" + _field_cparser[f](kwargs[f]) for f in kwargs]
		if not f: return u""
		return u" ".join([u""] + f)

	def modify_post(self, md5, **kwargs):
		fspec = self._fieldspec(**kwargs)
		if fspec:
			cmd = u"MP" + _uniw(md5) + fspec
			self._writeline(cmd)
			res = self._readline()
			if res != u"OK": raise ResponseError(res)

	def add_post(self, md5, **kwargs):
		cmd  = u"AP" + _uniw(md5)
		assert "width" in kwargs
		assert "height" in kwargs
		assert "ext" in kwargs
		cmd += self._fieldspec(**kwargs)
		self._writeline(cmd)
		res = self._readline()
		if res != u"OK": raise ResponseError(res)

	def _rels(self, c, md5, rels):
		cmd = [u"R", c, md5]
		for rel in self._list(rels, _uniw):
			cmd.append(u" " + rel)
		self._writeline(u"".join(cmd))
		res = self._readline()
		if res != u"OK": raise ResponseError(res)

	def add_rels(self, md5, rels):
		self._rels(u"R", md5, rels)

	def remove_rels(self, md5, rels):
		self._rels(u"r", md5, rels)

	def _parse_rels(self, line, rels):
		if line == u"OK": return True
		if line[0] != u"R": raise ResponseError(line)
		if line[1] == u"E": raise ResponseError(line)
		a = str(line[1:]).split()
		p = a[0]
		l = rels.setdefault(p, [])
		l += a[1:]

	def post_rels(self, md5):
		cmd = u"RS" + _uniw(md5)
		self._writeline(cmd)
		rels = {}
		while not self._parse_rels(self._readline(), rels): pass
		if not md5 in rels: return None
		return rels[md5]

	def add_tag(self, name, type=None, guid=None, valuetype=None):
		cmd = u"ATN" + _uniw(name)
		if type:
			cmd += u" T" + _uniw(type)
		if guid:
			cmd += u" G" + _uniw(guid)
		if valuetype:
			cmd += u" V" + _uniw(valuetype)
		self._writeline(cmd)
		res = self._readline()
		if res != u"OK": raise ResponseError(res)

	def add_alias(self, name, origin_guid):
		cmd = u"AAG" + _uniw(origin_guid) + u" N" + _uniw(name)
		self._writeline(cmd)
		res = self._readline()
		if res != u"OK": raise ResponseError(res)

	def remove_alias(self, name):
		cmd = u"DAN" + _uniw(name)
		self._writeline(cmd)
		res = self._readline()
		if res != u"OK": raise ResponseError(res)

	def _addrem_implies(self, addrem, set_tag, implied_tag, datastr, filter, value=None):
		set_tag = self._tag2spec(set_tag)
		implied_tag = self._tag2spec(implied_tag)
		assert u" " not in set_tag
		assert u" " not in implied_tag
		if implied_tag[0] == u"-":
			add = u" i" + implied_tag[1:]
		else:
			add = u" I" + implied_tag
		if filter:
			assert isinstance(filter, tuple) and len(filter) == 2
			filter = _uniw(filter[0]) + filter[1].format()
			assert u" " not in filter
			set_tag += filter
		if value:
			datastr += " V" + value.format()
		cmd = u"I" + addrem + set_tag + add + datastr
		self._writeline(cmd)
		res = self._readline()
		if res != u"OK": raise ResponseError(res)

	def add_implies(self, set_tag, implied_tag, priority=0, filter=None, value=None):
		datastr = u""
		if priority:
			datastr += u" P%d" % (priority,)
		self._addrem_implies("I", set_tag, implied_tag, datastr, filter, value)

	def remove_implies(self, set_tag, implied_tag, filter=None):
		self._addrem_implies("i", set_tag, implied_tag, "", filter)

	def _parse_implies(self, data):
		res = self._readline()
		if res == u"OK": return
		guid, impl = res.split(u" ", 1)
		assert guid[:2] == u"RI"
		guid = guid[2:]
		if len(guid) == 27:
			filter = None
		else:
			pt = self._parse_tag("", guid, 27, True, True)
			filter = pt[1:]
			if filter == (None, None): filter = None
			guid = pt[0]
		l = data.setdefault(guid, [])
		guid = None
		prio = 0
		value = None
		for part in impl.split():
			if part[0] in u"Ii":
				if guid:
					l.append(ImplicationTuple(guid, prio, filter, value))
				prio = 0
				value = None
				guid = str(part[1:])
				if part[0] == u"i":
					guid = "-" + guid
			elif part[0] == u"P":
				assert guid
				prio = int(part[1:])
			elif part[0] == u"V":
				assert guid
				value = part[1:]
				if value:
					valuetype, value = value.split("=", 1)
					value = vtparse(valuetype, value)
				else:
					value = InheritValue
			else:
				raise ResponseError(res)
		l.append(ImplicationTuple(guid, prio, filter, value))
		return True

	def tag_implies(self, tag, reverse=False):
		tag = _uniw(tag)
		cmd = u"IR" if reverse else u"IS"
		self._writeline(cmd + tag)
		data = {}
		while self._parse_implies(data):
			pass
		if reverse:
			rev = []
			for itag in data:
				for impl in data[itag]:
					assert len(impl) >= 2
					ttag = impl[0]
					if ttag[0] == "-":
						assert ttag[1:] == tag
						rev.append(impl._replace(guid="-" + itag))
					else:
						assert ttag == tag
						rev.append(impl._replace(guid=itag))
			return rev
		if tag in data:
			return data[tag]
		return []

	def merge_tags(self, into_t, from_t):
		cmd = u"MTG" + _uniw(into_t) + u" M" + _uniw(from_t)
		self._writeline(cmd)
		res = self._readline()
		if res != u"OK": raise ResponseError(res)

	def mod_tag(self, guid, name=None, type=None, valuetype=None):
		cmd = u"MTG" + _uniw(guid)
		for init, field in ((u" N", name), (u" T", type), (u" V", valuetype)):
			if field:
				cmd += init + _uniw(field)
		self._writeline(cmd)
		res = self._readline()
		if res != u"OK": raise ResponseError(res)

	def _tag2spec(self, t, value_allowed=True):
		if type(t) in (tuple, list):
			assert len(t) in (2, 3)
			g = _uniw(t[0])
			if t[-1] is None: return g
			assert value_allowed
			if len(t) == 2:
				return g + u"=" + t[1].format()
			else:
				return g + t[1] + t[2].format()
		else:
			return _uniw(t)

	def tag_post(self, md5, full_tags=None, weak_tags=None, remove_tags=None):
		tags = self._list(full_tags, lambda s: u" T" + self._tag2spec(s))
		tags += self._list(weak_tags, lambda s: u" T~" + self._tag2spec(s))
		tags += self._list(remove_tags, lambda s: u" t" + self._tag2spec(s, False))
		init = u"TP" + _uniw(md5)
		cmd = [init]
		clen = len(init)
		for tag in tags:
			assert u" " not in tag[1:]
			clen += len(tag.encode("utf-8"))
			if clen >= self._prot_max_len:
				self._writeline(u"".join(cmd))
				res = self._readline()
				if res != u"OK": raise ResponseError(res)
				cmd = [init]
				clen = len(init) + len(tag.encode("utf-8"))
			cmd.append(tag)
		if len(cmd) > 1:
			self._writeline(u"".join(cmd))
			res = self._readline()
			if res != u"OK": raise ResponseError(res)

	def _parse_tagres(self, resdata = None):
		res = self._readline()
		if res == u"OK": return
		if res[0] != u"R": raise ResponseError(res)
		if res[1] == u"E": raise ResponseError(res)
		if res[1] == u"R": return True # ignore count for now
		t = Tag()
		t._populate(res[1:])
		if resdata != None: resdata.append(t)
		return t

	def find_tags(self, matchtype, name, range=None, order=None, flags=None, **kw):
		if kw:
			filter = self._build_search(**kw)
			if filter:
				filter = u" :" + filter
		else:
			filter = u""
		matchtype = _uniw(matchtype)
		name = _uniw(name)
		cmd = [u"ST", matchtype, name]
		for o in self._list(order, _uniw):
			cmd.append(u" O" + o)
		for f in self._list(flags, _uniw):
			cmd.append(u" F" + f)
		if range is not None:
			assert len(range) == 2
			cmd.append(u" R%x:%x" % tuple(range))
		cmd.append(filter)
		self._writeline(u"".join(cmd))
		tags = []
		while self._parse_tagres(tags): pass
		return tags

	def _parse_tag(self, prefix, spec, pos, comparison, is_guid=False):
		if pos == -1:
			return None
		if is_guid:
			assert comparison
			ppos = -1
			tag = Tag()
			tag.guid = spec[:pos]
		else:
			tag = self.find_tag(spec[:pos])
			if comparison:
				ppos = _rfindany(spec, u"=<>", pos)
			else:
				ppos = spec.rfind(u"=", 0, pos)
			if not tag:
				return self._parse_tag(prefix, spec, ppos, comparison)
			tag = self.get_tag(tag)
			if not tag or tag.valuetype in (None, "none"):
				return self._parse_tag(prefix, spec, ppos, comparison)
		if comparison:
			if pos + 1 < len(spec) and spec[pos + 1] in u"=~":
				comp = spec[pos:pos + 2]
				val = spec[pos + 2:]
			else:
				comp = spec[pos]
				val = spec[pos + 1:]
			if is_guid:
				tag.valuetype, val = val.split("=", 1)
			if comp not in (u"=", u"<", u">", u"<=", u">=", u"=~"):
				return None
			val = vtparse(tag.valuetype, val, True)
			if comp != u"=" and isinstance(val, VTnull):
				return None
			return (prefix + tag.guid, comp, val)
		else:
			val = spec[pos + 1:]
			if val:
				val = vtparse(tag.valuetype, val, True)
			else:
				val = None
			return (prefix + tag.guid, val)

	def parse_tag(self, spec, comparison=False):
		spec = _uni(spec)
		if not spec: return None
		if spec[0] in u"~-!":
			prefix = spec[0]
			spec = spec[1:]
		else:
			prefix = u""
		if u" " not in spec and u"\n" not in spec:
			tag = self.find_tag(spec)
		else:
			tag = None
		if tag:
			if comparison:
				return (prefix + tag, None, None)
			else:
				return (prefix + tag, None)
		if comparison:
			ppos = _rfindany(spec, u"=<>", spec.find(u" "))
		else:
			ppos = spec.rfind(u"=", 0, spec.find(u" "))
		return self._parse_tag(prefix, spec, ppos, comparison)

	def _find_tag(self, matchtype, name, with_prefix):
		name = _uniw(name)
		if with_prefix and name[0] in u"~-!":
			prefix = name[0]
			name = name[1:]
		else:
			prefix = u""
		tags = self.find_tags(matchtype, name)
		if not tags: return None
		assert len(tags) == 1
		tag = tags[0]
		tag.pname = prefix + tag.name
		tag.pguid = prefix + tag.guid
		return tag

	def find_tag(self, name, resdata=None, with_prefix=False):
		tag = self._find_tag(u"EAN", name, with_prefix)
		if not tag: return None
		if resdata != None: resdata.update(tag)
		return tag.pguid

	def get_tag(self, guid, with_prefix=False):
		tag = self._find_tag(u"EAG", guid, with_prefix)
		if not tag: return None
		assert guid[-27:] == tag.guid
		return tag

	def begin_transaction(self):
		self._writeline(u"tB")
		res = self._readline()
		return res == u"OK"

	def end_transaction(self):
		self._writeline(u"tE")
		res = self._readline()
		return res == u"OK"

	def thumb_path(self, md5, size):
		md5 = str(md5)
		return os.path.join(self.cfg.thumb_base, str(size), md5[0], md5[1:3], md5)

	def pngthumb_path(self, md5, ft, size):
		fn = _uniw(md5) + u"." + _uniw(ft)
		md5 = hashlib.md5(fn.encode("utf-8")).hexdigest()
		return os.path.join(self.cfg.thumb_base, str(size), md5[0], md5[1:3], md5)

	def image_dir(self, md5):
		md5 = str(md5)
		return os.path.join(self.cfg.image_base, md5[0], md5[1:3])

	def image_path(self, md5):
		md5 = str(md5)
		return os.path.join(self.image_dir(md5), md5)

	def postspec2md5(self, spec, default=None):
		if os.path.lexists(spec) and not os.path.isdir(spec):
			# some extra magic to avoid reading the files if possible
			if os.path.islink(spec):
				dest = os.readlink(spec)
				m = self._destmd5re.match(dest)
				if m: return m.group(1)
			# Even when the fuse fs returns files, bare IDs are links
			aspec = spec.split("/")
			afn = aspec[-1].split(".")[-2:]
			if self._md5re.match(afn[0]):
				aspec[-1] = afn[0]
				shortspec = "/".join(aspec)
				if os.path.islink(shortspec):
					dest = os.readlink(shortspec)
					m = self._destmd5re.match(dest)
					if m: return m.group(1)
			# Oh well, hash the file.
			return hashlib.md5(open(spec, "rb").read()).hexdigest()
		if self._md5re.match(spec): return spec
		return default

	def order(self, tag, posts):
		init = u"OG" + _uniw(tag)
		cmd = [init]
		dolen = 2
		clen = len(init)
		for post in map(_uniw, posts):
			post = u" P" + post
			cmd.append(post)
			clen += len(post.encode("utf-8"))
			if clen >= self._prot_max_len - 35:
				self._writeline(u"".join(cmd))
				res = self._readline()
				if res != u"OK": raise ResponseError(res)
				# Overlap one, so previous ordering doesn't mess us up.
				cmd = [init, post]
				dolen = 3
				clen = len(init) + len(post.encode("utf-8"))
		if len(cmd) >= dolen:
			self._writeline(u"".join(cmd))
			res = self._readline()
			if res != u"OK": raise ResponseError(res)

	def metalist(self, name):
		cmd = u"L" + _uniw(name)
		self._writeline(cmd)
		res = self._readline()
		names = []
		while res != u"OK":
			if res[:2] != u"RN": raise ResponseError(res)
			names.append(res[2:])
			res = self._readline()
		return names

	def thumb_fns(self, m, ft):
		sizes = self.cfg.thumb_sizes.split()
		jpeg_fns = map(lambda z: (self.thumb_path(m, int(z)), int(z)), sizes)
		png_fns = map(lambda n, z: (self.pngthumb_path(m, ft, n), z),
		              ("normal", "large"), (128, 256))
		return list(jpeg_fns), list(png_fns)

	def save_thumbs(self, m, img, ft, rot, force=False):
		from PIL import Image, PngImagePlugin
		fn = self.image_path(m)
		mtime = os.stat(fn).st_mtime
		if not img:
			img = Image.open(RawWrapper(open(fn, "rb")))
		img.load()
		# PIL rotates CCW
		rotation = {90: Image.ROTATE_270, 180: Image.ROTATE_180, 270: Image.ROTATE_90}
		if rot in rotation: img = img.transpose(rotation[rot])
		w, h = img.size
		if img.mode == "1":
			# We want to scale B/W as grayscale.
			img = img.convert("L")
		if img.mode == "P" and "transparency" in img.info:
			# special case for transparent gif
			img = img.convert("RGBA")
		if img.mode not in ("RGB", "RGBA", "L", "LA"):
			# Are there other modes to worry about?
			img = img.convert("RGB")
		jpeg_fns, png_fns = self.thumb_fns(m, ft)
		jpeg_opts = {"format": "JPEG", "quality": 95, "optimize": 1}
		meta = PngImagePlugin.PngInfo()
		meta.add_text("Thumb::URI", str(m + "." + ft), 0)
		meta.add_text("Thumb::MTime", str(int(mtime)), 0)
		png_opts = {"format": "PNG", "pnginfo": meta}
		jpeg = list(map(lambda t: (t[0], t[1], jpeg_opts), jpeg_fns))
		png = list(map(lambda t: (t[0], t[1], png_opts), png_fns))
		z = max(map(lambda d: d[1], jpeg + png)) * 2
		if w > z or h > z:
			img = _thumb(img, z)
		for fn, z, opts in jpeg + png:
			if force or not os.path.exists(fn):
				t = _thumb(img.copy(), z)
				make_pdirs(fn)
				if opts["format"] == "JPEG":
					# Some versions of PIL care, some don't.
					if t.mode == "RGBA":
						t = t.convert("RGB")
					if t.mode == "LA":
						t = t.convert("L")
				t.save(fn, **opts)

def _thumb(img, z):
	from PIL import Image
	if max(img.size) > z:
		img.thumbnail((z, z), Image.ANTIALIAS)
	if img.mode[-1] == "A":
		# Images with transparency tend to have crap in the
		# tansparent pixel values. In some versions of PIL
		# this reappears after scaling.
		# So we put it on a white background every time.
		if img.mode == "LA":
			mode = "LA"
			col = 255
		else:
			mode = "RGBA"
			col = (255, 255, 255)
		bgfix = Image.new(mode, img.size, col)
		alpha = img.split()[-1]
		bgfix.paste(img, None, alpha)
		bgfix.putalpha(alpha)
		return bgfix
	else:
		return img



from __future__ import print_function

from wellpapp import Client

def main(arg0, argv):
	if len(argv) < 2:
		print("Usage:", arg0, "tagname post-spec [post-spec [...]]")
		return 1

	client = Client()
	data = {}
	tag = client.find_tag(argv[0], data)
	if not tag:
		print("Tag not found")
		return 1
	if data["weak_posts"]:
		print("Can't order a tag with weak posts")
		return 1
	posts = map(client.postspec2md5, argv[1:])
	wtag = "~" + tag
	for post, spec in zip(posts, argv[1:]):
		if post:
			post = client.get_post(post, True)
		if not post:
			print("Post " + spec + " not found")
			return 1
		if tag in post.implfulltags or tag in post.implweaktags:
			print("Post " + spec + " has tag implied.")
			print("Can't order a tag with implied posts.")
			return 1
		if tag not in post.tags.guids:
			print("Post " + spec + " doesn't have tag.")
			return 1
	client.order(tag, posts)



from __future__ import print_function

from sys import version_info
from wellpapp import Client
from re import match
from os import readlink, stat
from optparse import OptionParser

if version_info[0] > 2:
	unicode = str

def main(arg0, argv):
	def implfmt(impl):
		data = client.get_tag(impl.guid, with_prefix=True)
		a = [data.pname, unicode(impl.prio)]
		if impl.filter:
			a.append(impl.filter[0] + unicode(impl.filter[1]))
			a.append(unicode(impl.value))
		elif impl.value:
			a.append(u":")
			a.append(unicode(impl.value))
		return u"\n\t" + u"\t".join(a)

	def show_implies(guid, heading, reverse):
		impl = client.tag_implies(guid, reverse)
		if impl: print(heading + u"".join(sorted(map(implfmt, impl))))

	def fmt_tag(tag):
		if tag.value:
			val = u" = " + unicode(tag.value)
		else:
			val = u""
		return tag.pname + val

	def show_post(m, short=False, show_thumbs=False):
		post = client.get_post(m, True, ["tagname", "tagdata", "datatags", "ext", "created", "width", "height"])
		if not post:
			print("Post not found")
			return 1
		print(m + " created " + post.created.localtimestr())
		if not short:
			print(post["width"], "x", post["height"], post["ext"])
		try:
			path = readlink(client.image_path(m))
			try:
				path += " (%d bytes)" % (stat(path).st_size,)
			except OSError:
				path += " (MISSING)"
		except Exception:
			path = "MISSING"
		print("Original file: " + path)
		if show_thumbs:
			jpeg, png = client.thumb_fns(m, post.ext)
			for kind, lst in (("jpeg", jpeg), ("png", png)):
				for fn, z in lst:
					try:
						extra = "%d bytes" % (stat(fn).st_size,)
					except OSError:
						extra = "MISSING"
					print("Thumb %dpx %s: %s (%s)" % (z, kind, fn, extra))
		if short: return 0
		tags = [fmt_tag(t) for n, t in sorted(post.datatags.items())]
		if tags:
			print("Data:\n\t", end="")
			print(u"\n\t".join(sorted(tags)))
		print("Tags:\n\t", end="")
		tags = [fmt_tag(t) for t in post.fulltags] + [fmt_tag(t) for t in post.weaktags]
		print(u"\n\t".join(sorted(tags)))
		tags = [fmt_tag(t) for t in post.implfulltags] + [fmt_tag(t) for t in post.implweaktags]
		if tags:
			print("Implied:\n\t", end="")
			print(u"\n\t".join(sorted(tags)))
		rels = client.post_rels(m)
		if rels:
			print("Related posts:\n\t" + "\n\t".join(rels))
		return 0

	def show_tag(name, short=False, show_thumbs=False):
		guid = client.find_tag(name)
		if not guid and match(r"(?:\w{6}-){3}\w{6}", name):
			guid = name
		if not guid:
			print("Tag not found")
			return 1
		data = client.get_tag(guid)
		if not data:
			print("Tag not found")
			return 1
		print("Tag:", data["name"])
		if data.alias:
			if len(data.alias) == 1:
				print("Aliases:", data.alias[0])
			else:
				print("Aliases:")
				for a in sorted(data.alias):
					print("\t" + a)
		print("GUID:", guid)
		print("Type:", data["type"])
		if "valuetype" in data and data["valuetype"]:
			print("Valuetype:", data["valuetype"])
		if short: return 0
		print(data["posts"], "posts")
		print(data["weak_posts"], "weak posts")
		show_implies(guid, u"Implies:", False)
		show_implies(guid, u"Implied by:", True)
		flags = [f for f in data if data[f] is True]
		if flags:
			print("Flags:\n\t" + "\n\t".join(flags))
		return 0

	p = OptionParser(usage="Usage: %prog [-qt] post-spec or tagname [...]", prog=arg0)
	p.add_option("-q", "--short",
	             action="store_true",
	             help="Short output format"
	            )
	p.add_option("-t", "--show-thumbs",
	             action="store_true",
	             help="Show thumb paths"
	            )
	opts, args = p.parse_args(argv)
	if not args:
		p.print_help()
		return 1
	client = Client()
	ret = 0
	for object in args:
		object = client.postspec2md5(object, object)
		if match(r"^[0-9a-f]{32}$", object):
			ret |= show_post(object, **opts.__dict__)
		else:
			ret |= show_tag(object, **opts.__dict__)
		if len(args) > 1: print()
	return ret



from __future__ import print_function

from wellpapp import Client

def main(arg0, argv):
	if len(argv) not in (2, 3, 4):
		print("Usage:", arg0, "tag new_name [new_type [new_valuetype]]")
		return 1

	a = {}
	if len(argv) > 2:
		a["type"] = argv[2]
	if len(argv) > 3:
		a["valuetype"] = argv[3]

	client = Client()
	tag = client.find_tag(argv[0])
	a["name"] = argv[1]
	if not tag:
		print("Tag not found")
		return 1
	client.mod_tag(tag, **a)



from __future__ import print_function

from wellpapp import Client

def main(arg0, argv):
	if len(argv) < 2:
		print("Usage:", arg0, "post-spec property=value [property=value [...]]")
		print("or:", arg0, "-r property=value post-spec [post-spec [...]]")
		print("Properties as specified in decimal format where possible.")
		print("(So width=42, not width=2a)")
		return 1

	def set_prop(props, spec):
		prop, val = spec.split("=", 1)
		props[prop] = val

	client = Client()
	props = {}
	if argv[0] == "-r":
		set_prop(props, argv[1])
		client.begin_transaction()
		for post in argv[2:]:
			md5 = client.postspec2md5(post)
			if not md5 or not client.get_post(md5):
				print(post, "not found")
				continue
			try:
				client.modify_post(md5, **props)
			except Exception:
				print("Failed to set on", post)
		client.end_transaction()
	else:
		md5 = client.postspec2md5(argv[0])
		if not md5 or not client.get_post(md5):
			print("Post not found")
			return 1
		for prop in argv[1:]:
			set_prop(props, prop)
		if props:
			client.modify_post(md5, **props)



from __future__ import print_function

from wellpapp import Client

def main(arg0, argv):
	if len(argv) != 2:
		print("Usage:", arg0, "tagname alias")
		return 1
	client = Client()
	tag = client.find_tag(argv[0])
	if tag:
		client.add_alias(argv[1], tag)
	else:
		print("No such tag", argv[0])
		return 1



from __future__ import print_function
from __future__ import unicode_literals

from hashlib import md5
from PIL import Image
from io import BytesIO
from os.path import basename, dirname, realpath, exists, lexists, join, sep
from os import readlink, symlink, unlink, stat
from wellpapp import Client, VTstring, make_pdirs, RawWrapper, identify_raw, ExifWrapper, raw_exts, VTgps
from struct import unpack
from multiprocessing import Lock, cpu_count, Queue, Process
from traceback import print_exc
from sys import version_info, exit

if version_info[0] == 2:
	from Queue import Empty
	from io import open
else:
	from queue import Empty

def determine_filetype(data):
	data2 = data[:2]
	data3 = data[:3]
	data4 = data[:4]
	data6 = data[:6]
	data8 = data[:8]
	data16 = data[:16]
	data48 = data[4:8]
	if data3 == b"\xff\xd8\xff":
		# probably jpeg, but I like to be careful.
		if data[-2:] == b"\xff\xd9":
			# this is how a jpeg should end
			return "jpeg"
		if data[-4:] == b"SEFT" and b"SEFH" in data[-100:] and b"\xff\xd9\x00\x00" in data[-256:]:
			# samsung phones like to add this crap after the jpeg.
			# no idea why it's not in a makernote, but here we are.
			return "jpeg"
	if data6 in (b"GIF87a", b"GIF89a") and data[-1:] == b";":
		return "gif"
	if data8 == b"\x89PNG\r\n\x1a\n" and data[12:16] == b"IHDR":
		return "png"
	if data2 == b"BM" and ord(data[5:6]) < 4 and data[13:14] == b"\x00":
		return "bmp"
	if data4 in (b"MM\x00*", b"II*\x00", b"IIRO", b"IIU\x00", b"FOVb", b"FUJI",):
		return identify_raw(BytesIO(data))
	flen = unpack("<I", data48)[0]
	dlen = len(data)
	if data3 == b"FWS" and flen == dlen:
		return "swf"
	if data3 == b"CWS" and dlen < flen < dlen * 10:
		return "swf"
	if data4 == b"RIFF" and data[8:12] == b"AVI " and flen < dlen < flen * 1.4 + 10240:
		return "avi"
	if data3 == b"\x00\x00\x01":
		return "mpeg"
	if data4 == b"\x1a\x45\xdf\xa3" and b"matroska" in data[:64]:
		return "mkv"
	if data4 == b"\x1a\x45\xdf\xa3" and b"webm" in data[:64]:
		return "webm"
	if data4 == b"OggS" and data[28:35] in (b"\x01video\x00", b"\x80theora"):
		return "ogm"
	if data16 == b"\x30\x26\xb2\x75\x8e\x66\xcf\x11\xa6\xd9\x00\xaa\x00\x62\xce\x6c":
		return "wmv"
	if data48 in (b"ftyp", b"mdat"):
		blen = unpack(">I", data4)[0]
		if data[blen + 4:blen + 8] == b"moov":
			if data48 == b"mdat" or data[8:10] == b"qt":
				return "mov"
			else:
				return "mp4"
		if blen < 100 and (data[8:10] == b"qt" or data[8:12] in (b"3gp4", b"\0\0\0\0")):
			return "mov"
		if data[0:1] == b"\x00" and data[8:12] in (b"mp41", b"mp42", b"isom"):
			return "mp4"
	if data48 == b"moov" and data[12:16] in (b"mvhd", b"cmov"):
		if b"\x00" in (data[0:1], data[8:9], data[16:17]):
			return "mov"
	if data4 == b"FLV\x01" and data[5:9] == b"\x00\x00\x00\x09" and ord(data[13:14]) in (8, 9, 18):
		return "flv"
	if data.startswith(b"%PDF-"):
		return "pdf"

def _gpspos(pos, ref):
	pos = pos[0] + pos[1] / 60 + pos[2] / 3600
	if ref and ref[0] in "SWsw":
		pos = -pos
	return "%.7f" % (pos,)

def fmt_tagvalue(v):
	if not v: return ""
	if isinstance(v, VTstring):
		return "=" + repr(v.str)
	else:
		return "=" + v.str

def flash_dimensions(data):
	sig = data[:3]
	assert sig in (b"FWS", b"CWS")
	data = data[8:]
	if sig == b"CWS":
		from zlib import decompress
		data = decompress(data)
	pos = [0] # No "nonlocal" in python2
	def get_bits(n, signed=True):
		val = 0
		for i in range(n):
			sh = 7 - (pos[0] % 8)
			bit = (ord(data[pos[0] // 8]) & (1 << sh)) >> sh
			if i == 0 and signed and bit:
				val = -1
			val = (val << 1) | bit
			pos[0] += 1
		return val
	bpv = get_bits(5, False)
	xmin = get_bits(bpv)
	xmax = get_bits(bpv)
	ymin = get_bits(bpv)
	ymax = get_bits(bpv)
	return [int(round(v / 20.0)) for v in (xmax - xmin, ymax - ymin)]

def mplayer_dimensions(fn):
	from subprocess import Popen, PIPE, STDOUT
	p = Popen(["mediainfo", "--Inform=Video;%Width% %Height%", fn],
	           stdout=PIPE, stderr=STDOUT, close_fds=True)
	data = p.communicate()[0].split()
	return tuple(map(int, data))

movie_ft = set("swf avi mpeg mkv ogm mp4 wmv flv mov webm".split())

def pdf_image(fn):
	from subprocess import check_output
	cmd = ["gs",
	       "-q",
	       "-sDEVICE=pngalpha",
	       "-sOutputFile=-",
	       "-dFirstPage=1",
	       "-dLastPage=1",
	       "-dBATCH",
	       "-dNOPAUSE",
	       "-dSAFER",
	       "-r100",
	       fn,
	]
	return check_output(cmd, close_fds=True)


def main(arg0, argv):
	def needs_thumbs(m, ft):
		if force_thumbs: return True
		jpeg_fns, png_fns = client.thumb_fns(m, ft)
		for fn, z in jpeg_fns + png_fns:
			if not exists(fn): return True

	def exif2tags(exif, tags):
		cfg = client.cfg
		if "lenstags" in cfg:
			lenstags = cfg.lenstags.split()
			for lt in lenstags:
				if lt in exif:
					v = exif[lt]
					if isinstance(v, tuple) or hasattr(v, "pop"):
						v = " ".join([str(e) for e in v])
					lt = "lens:" + lt + ":" + v
					if lt in cfg:
						tags.add_spec(cfg[lt])
		try:
			make = exif["Exif.Image.Make"].strip()
			model = exif["Exif.Image.Model"].strip()
			cam = "camera:" + make + ":" + model
			if cam in cfg:
				tags.add_spec(cfg[cam])
		except Exception:
			pass
		if "set_tags" in cfg:
			for st in cfg.set_tags.split():
				tn, et = st.split("=", 1)
				if et in exif:
					v = exif[et]
					if ("FocalLength" in et or "FNumber" in et) and not v:
						continue
					tags.add_spec(tn + "=" + str(exif[et]))
		if "Exif.GPSInfo.GPSLatitude" in exif:
			lat = _gpspos(exif["Exif.GPSInfo.GPSLatitude"], exif["Exif.GPSInfo.GPSLatitudeRef"])
			lon = _gpspos(exif["Exif.GPSInfo.GPSLongitude"], exif["Exif.GPSInfo.GPSLongitudeRef"])
			if "Exif.GPSInfo.GPSAltitude" in exif:
				from fractions import Fraction
				alt = Fraction(exif["Exif.GPSInfo.GPSAltitude"])
				if "Exif.GPSInfo.GPSAltitudeRef" in exif and float(exif["Exif.GPSInfo.GPSAltitudeRef"]):
					alt = -alt
				gps = VTgps("%s,%s,%.1f" % (lat, lon, alt))
			else:
				gps = VTgps("%s,%s" % (lat, lon))
			tags.add(("aaaaaa-aaaadt-faketg-gpspos", gps))
		if override_tags:
			tags.add(("aaaaaa-aaaac8-faketg-bddate", exif.date(tz)))

	class tagset(set):
		def add(self, t):
			guid, val = t
			prefix = ""
			if guid[0] in "~-":
				prefix = guid[0]
				guid = guid[1:]
			chk = (guid, "~" + guid)
			rem = None
			for v in self:
				if v[0] in chk: rem = v
			if rem: self.remove(rem)
			if prefix != "-": set.add(self, (prefix + guid, val))

		def difference(self, other):
			other = {(guid, None if val is None else val.format()) for guid, val in other}
			for guid, val in self:
				if (guid, None if val is None else val.format()) not in other:
					yield guid, val

		def add_spec(self, s):
			try:
				with lock:
					t = client.parse_tag(s)
			except Exception:
				print("Failed to parse: " + s)
				return
			if t:
				self.add(t)
			else:
				print("Unknown tag " + s)

		def update(self, l):
			[self.add_spec(s) for s in l]

		def update_tags(self, l):
			[self.add((t.pguid, t.value)) for t in l]

	def find_tags(fn):
		path = "/"
		tags = tagset()
		for dir in dirname(fn).split(sep):
			path = join(path, dir)
			TAGS = join(path, "TAGS")
			if exists(TAGS):
				tags.update(open(TAGS, "r", encoding="utf-8").readline().split())
		if gen_from_fn:
			tags.update(basename(fn).split()[:-1])
		return tags

	def record_filename(m, fn):
		dn = client.image_dir(m)
		rec_fn = join(dn, "FILENAMES")
		known = {}
		if exists(rec_fn):
			for line in open(rec_fn, "r", encoding="utf-8", errors="backslashreplace"):
				r_m, r_fn = line[:-1].split(" ", 1)
				known.setdefault(r_m, []).append(r_fn)
		if m not in known or fn not in known[m]:
			with lock:
				with open(rec_fn, "a", encoding="utf-8", errors="backslashreplace") as fh:
					fh.write(m + " " + fn + "\n")

	def generate_cache(m, fn, jz):
		cache_fn = client.cfg.image_base + "/cache"
		if exists(cache_fn):
			s = stat(fn)
			z = s.st_size
			mt = int(s.st_mtime)
			if jz:
				l = "1 %s %d %d %d %s\n" % (m, z, mt, jz, fn)
			else:
				l = "0 %s %d %d %s\n" % (m, z, mt, fn)
			with lock:
				with open(cache_fn, "a", encoding="utf-8") as fh:
					fh.write(l)

	def add_image(fn, warn_q):
		if verbose: print(fn)
		fn = realpath(fn)
		data = open(fn, "rb").read()
		m = md5(data).hexdigest()
		with lock:
			post = client.get_post(m, True)
		if post:
			ft = post.ext
		else:
			ft = force_filetype or determine_filetype(data)
		assert ft
		p = client.image_path(m)
		if lexists(p):
			ld = readlink(p)
			is_wpfs = False
			try:
				dot = fn.rindex(".")
				if "/" not in fn[dot:]:
					bare_fn = fn[:fn.rindex(".")]
					if m == bare_fn[-32:] and readlink(bare_fn)[-32:] == m:
						is_wpfs = True # probably
			except (OSError, ValueError):
				pass
			if is_wpfs:
				if not quiet:
					print("Not updating", fn, "because this looks like a wellpappfs")
			elif exists(p):
				if fn != ld:
					if not dummy: record_filename(m, fn)
					if not quiet:
						print("Not updating", m, fn)
			else:
				if dummy:
					if not quiet: print("Would have updated", m, fn)
				else:
					record_filename(m, ld)
					if not quiet: print("Updating", m, fn)
					unlink(p)
		do_cache = False
		if not lexists(p) and not dummy:
			make_pdirs(p)
			symlink(fn, p)
			do_cache = True
		if not post or needs_thumbs(m, ft):
			do_cache = True
			if ft in movie_ft:
				if not thumb_source:
					print("Can't generate " + ft + " thumbnails")
					exit(1)
				if not post:
					if ft == "swf":
						w, h = flash_dimensions(data)
					else:
						w, h = mplayer_dimensions(fn)
			else:
				if ft == "pdf":
					data = pdf_image(fn)
				datafh = RawWrapper(BytesIO(data))
				try:
					img = Image.open(datafh)
				except IOError:
					if thumb_source:
						img = Image.open(RawWrapper(open(thumb_source, "rb")))
						print("Warning: taking dimensions from thumb source")
					else:
						raise
				w, h = img.size
		if do_cache and not dummy:
			if ft in raw_exts:
				jfh = RawWrapper(BytesIO(data), True)
				jfh.seek(0, 2)
				jz = jfh.tell()
				jfh.close()
			else:
				jz = None
			generate_cache(m, fn, jz)
		exif = ExifWrapper(fn)
		if not post:
			rot = exif.rotation()
			if rot in (90, 270): w, h = h, w
			args = {"md5": m, "width": w, "height": h, "ext": ft}
			if rot >= 0: args["rotate"] = rot
			date = exif.date(tz)
			if date:
				args["imgdate"] = date
			if dummy:
				print("Would have created post " + m)
			else:
				with lock:
					client.add_post(**args)
		if needs_thumbs(m, ft):
			if dummy:
				print("Would have generated thumbs for " + m)
			else:
				rot = exif.rotation()
				if thumb_source:
					img = Image.open(RawWrapper(open(thumb_source, "rb")))
				client.save_thumbs(m, img, ft, rot, force_thumbs)
		full = tagset()
		weak = tagset()
		with lock:
			post = client.get_post(m, True)
		posttags = tagset()
		if post:
			posttags.update_tags(post.tags)
		filetags = find_tags(fn)
		try:
			exif2tags(exif, filetags)
		except Exception:
			print_exc()
			warn_q.put(fn + ": failed to set tags from exif")
		for guid, val in filetags.difference(posttags):
			if guid in post.tags and not override_tags:
				print("Not overriding value on", post.tags[guid].pname)
			elif guid[0] == "~":
				weak.add((guid[1:], val))
			else:
				full.add((guid, val))
		if full or weak:
			with lock:
				if no_tagging or dummy:
					full = [client.get_tag(g).name + fmt_tagvalue(v) for g, v in full]
					weak = ["~" + client.get_tag(g).name + fmt_tagvalue(v) for g, v in weak]
					print("Would have tagged " + m + " " + " ".join(full + weak))
				else:
					client.tag_post(m, full, weak)

	def usage():
		print("Usage:", arg0, "[-v] [-q] [-f] [-n] [-d] [-t thumb_source] [-T type] [-z timezone] filename [filename [..]]")
		print("\t-v Verbose")
		print("\t-q Quiet")
		print("\t-f Force thumbnail regeneration")
		print("\t-n No tagging (prints what would have been tagged)")
		print("\t-g Generate tags from filename (all words except last)")
		print("\t-d Dummy, only print what would be done")
		print("\t-t Post or file to generate thumb from")
		print("\t-T Override file type detection")
		print("\t-z Timezone to assume EXIF dates are in (+-HHMM format)")
		print("\t-o Override existing tag values (from exif, TAGS (and filename))")
		exit(1)

	if len(argv) < 1: usage()
	a = 0
	switches = ("-v", "-q", "-f", "-h", "-n", "-d", "-t", "-T", "-z", "-o")
	quiet = False
	verbose = False
	force_thumbs = False
	no_tagging = False
	gen_from_fn = False
	dummy = False
	thumb_source = None
	tz = None
	override_tags = False
	force_filetype = False
	while argv[a] in switches:
		if argv[a] == "-q":
			quiet = True
		elif argv[a] == "-v":
			verbose = True
		elif argv[a] == "-f":
			force_thumbs = True
		elif argv[a] == "-n":
			no_tagging = True
		elif argv[a] == "-g":
			gen_from_fn = True
		elif argv[a] == "-d":
			dummy = True
		elif argv[a] == "-t":
			a += 1
			if len(argv) == a: usage()
			thumb_source = argv[a]
			force_thumbs = True
		elif argv[a] == "-T":
			a += 1
			if len(argv) == a: usage()
			force_filetype = argv[a]
		elif argv[a] == "-z":
			a += 1
			if len(argv) == a: usage()
			tz = argv[a]
		elif argv[a] == "-o":
			override_tags = True
		else:
			usage()
		a += 1
		if len(argv) == a: usage()
	client = Client()
	lock = Lock()

	if thumb_source:
		if len(argv) > a + 1:
			print("Only add one post with -t")
			exit(1)
		if not exists(thumb_source):
			m = client.postspec2md5(thumb_source)
			thumb_source = client.image_path(m)
		if not exists(thumb_source):
			print("Thumb source not found")
			exit(1)
	todo = argv[a:]
	if not todo:
		exit(0)
	client.begin_transaction()
	q = Queue()
	bad_q = Queue()
	warn_q = Queue()
	for td in todo:
		q.put(td)
	def worker():
		while True:
			try:
				i = q.get(False)
			except Empty:
				break
			try:
				add_image(i, warn_q)
			except Exception:
				print_exc()
				bad_q.put(i)
	# I would have used Pool, but it's completely broken if you ^C
	ps = [Process(target=worker) for _ in range(min(cpu_count(), len(todo)))]
	for p in ps:
		p.daemon = True
		p.start()
	for p in ps:
		p.join()
	client.end_transaction()
	def print_q(q, msg):
		if not q.empty():
			print()
			print(msg)
			while True:
				try:
					print("\t%s" % (q.get(False),))
				except Empty:
					break
	print_q(warn_q, "Files with warnings:")
	print_q(bad_q, "Failed files:")



from __future__ import print_function

from sys import exit
from wellpapp import Client
from argparse import ArgumentParser
from operator import itemgetter

def main(arg0, argv):
	parser = ArgumentParser(prog=arg0)
	parser.add_argument("-f", "--fuzzy", action="store_true", help="Fuzzy matching")
	parser.add_argument("-a", "--anywhere", action="store_true", help="Match anywhere")
	parser.add_argument("-s", "--short", action="store_true", help="Short listing (just tagnames)")
	parser.add_argument("-i", "--implies", action="append", default=[], metavar="TAG", help="Only find tags that imply this tag")
	parser.add_argument("-t", "--type", action="append", default=[], help="Only tags of this type")
	parser.add_argument(metavar="partial-tag-name", dest="part")
	args = parser.parse_args(argv)

	client = Client()

	def short(tag):
		print(tag.name)

	def long(tag):
		names = [tag.name] + sorted(tag.alias)
		props = [str(p) for p in (tag.type, tag.posts, tag.weak_posts)]
		if tag.valuetype: props.insert(0, "valuetype:" + tag.valuetype)
		print(" ".join(names + props))

	def parse_tags(lst):
		for name in lst:
			guid = client.find_tag(name, with_prefix=True)
			if not guid:
				print("Tag %s not found" % (name,))
				exit(1)
			yield guid

	def collect_implies(guid, res):
		for impl in client.tag_implies(guid):
			if impl.guid not in res:
				res.add(impl.guid)
				if not impl.guid.startswith("-"):
					collect_implies(impl.guid, res)

	match = "F" if args.fuzzy else "E"
	where = "P" if args.anywhere else "I"
	cmd = match + "A" + where
	printer = short if args.short else long
	implies = set(parse_tags(args.implies))
	tagtypes = set(args.type)

	for tag in sorted(client.find_tags(cmd, args.part), key=itemgetter('name')):
		if implies:
			tag_implies = set()
			collect_implies(tag.guid, tag_implies)
			if implies - tag_implies:
				continue
		if tagtypes:
			if tag.type not in tagtypes:
				continue
		printer(tag)



#!/usr/bin/env python3

from setuptools import setup

fuse_reqs = [
	'fuse-python >= 0.3.1; python_version < "3"',
	'fuse-python >= 1.0.0; python_version > "3"',
]

readme = open('README.md', 'r').read()
readme = readme.replace(
	'(FUSE.md)',
	'(https://github.com/drougge/wellpapp-pyclient/blob/master/FUSE.md)'
)

setup(
	name='wellpapp',
	version='2021.1.24', # set this for each release

	packages=[
		'wellpapp',
		'wellpapp.shell',
	],
	entry_points={
		'console_scripts': [
			'wp = wellpapp.__main__:main',
		],
	},
	install_requires=[
		'Pillow >= 3.1.2',
		'PyGObject >= 3.20',
	],
	extras_require={
		'fuse': fuse_reqs,
		'all': fuse_reqs,
	},
	python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',

	author='Carl Drougge',
	author_email='bearded@longhaired.org',
	url='https://github.com/drougge/wellpapp-pyclient',
	license='MIT',
	description='Client library and application for the wellpapp image tagging system.',
	long_description=readme,
	long_description_content_type='text/markdown',
)



from __future__ import print_function

from hashlib import md5
from PIL import Image
from os.path import exists
from os import stat, walk
from sys import version_info, exit
from wellpapp import Client

if version_info[0] > 2:
	basestring = str

def main(arg0, argv):
	def usage():
		print("Usage:", arg0, "-opts")
		print("Where opts can be:")
		print("\t-t Check thumb existance")
		print("\t-T Check thumb integrity (decodeability and timestamp)")
		print("\t-i Check image existance")
		print("\t-I Check image integrity (md5)")
		print("\t-f Check that all existing symlinks/thumbnails have a corresponding post")
		print("\t-s Check that all posts have the same tags they are findable by.")
		print("\t-a All of the above")
		exit(1)

	optchars = "tTiIfs"
	opts = ""
	for a in argv:
		if a[0] != "-": usage()
		for c in a[1:]:
			if c in optchars and c not in opts:
				opts += c
			elif c == "a":
				opts = optchars
			else:
				usage()
	if not opts: usage()

	def check_images(integrity):
		print("Checking images")
		bad = set()
		for m in posts:
			path = client.image_path(m)
			if not exists(path):
				print("Missing image", m)
				bad.add(m)
				continue
			if integrity:
				if md5(open(path, "rb").read()).hexdigest() != m:
					print("Bad file", m)
					bad.add(m)
		return bad

	def check_thumbs(integrity):
		print("Checking thumbs")
		sizes = list(map(int, client.cfg.thumb_sizes.split())) + ["normal", "large"]
		bad = set()
		for m in posts:
			thumbs = []
			for z in sizes:
				if isinstance(z, int):
					path = client.thumb_path(m, z)
				else:
					path = client.pngthumb_path(m, posts[m]["ext"], z)
				thumbs.append((path, z))
			for path, z in thumbs:
				if not exists(path):
					print(m, "missing thumb", z)
					bad.add(m)
			if integrity and m not in bad:
				for path, z in thumbs:
					img = Image.open(path)
					try:
						# img.verify() doesn't work on jpeg
						# (and load() is pretty forgiving too, but that's
						#  probably a problem with the format.)
						img.load()
					except Exception:
						print(m, "bad thumb", z)
						bad.add(m)
						continue
					if isinstance(z, basestring):
						if "Thumb::URI" not in img.info or "Thumb::MTime" not in img.info:
							print(m, "bad thumb", z)
							bad.add(m)
							continue
						t_mtime = int(img.info["Thumb::MTime"])
						f_mtime = int(stat(client.image_path(m)).st_mtime)
						if t_mtime != f_mtime:
							print(m, "outdated thumb", z)
							bad.add(m)
							continue
		return bad

	def _check_imagestore(msg, bad, dp, fns, name2md5):
		initial = dp[-4] + dp[-2:]
		for fn in fns:
			p = dp + "/" + fn
			if fn[:3] != initial:
				print("Misplaced", msg, p)
				bad.add(p)
				continue
			if name2md5(p)[-32:] not in posts:
				print(msg.title(), "without post", fn)
				bad.add(p)

	def _pngthumb2md5(fn):
		img = Image.open(fn)
		return img.info["Thumb::URI"][:32]

	def check_imagestore():
		bad = set()
		print("Checking for stray images")
		for dp, dns, fns in walk(client.cfg.image_base):
			_check_imagestore("image", bad, dp, fns, str)
		print("Checking for stray thumbnails")
		jpegz = client.cfg.thumb_sizes.split()
		for z, name2md5 in zip(jpegz + ["normal", "large"], [str] * len(jpegz) + [_pngthumb2md5] * 2):
			print("  " + z)
			for dp, dns, fns in walk(client.cfg.thumb_base + "/" + z):
				_check_imagestore("thumb", bad, dp, fns, name2md5)
		return bad

	def check_connectivity():
		print("Checking tags")
		for t in tags:
			strong = client.search_post(guids=["!" + t.guid])
			weak = client.search_post(guids=["~" + t.guid])
			if len(strong) != t.posts or len(weak) != t.weak_posts:
				print("Post count mismatch on", t.guid)
				print("\tclaims", t.posts, "+", t.weak_posts)
				print("\tfinds", len(strong), "+", len(weak))
			for res, prefix in ((strong, ""), (weak, "~")):
				for p in res:
					p = posts[p.md5]
					if prefix + t.guid not in p.tags.guids:
						print(p["md5"], "reachable with", t.guid, "but not tagged")
		print("Checking posts")
		for m, p in posts.items():
			guids = ["!" + t.guid for t in p.fulltags] + ["~" + t.guid for t in p.weaktags]
			while guids:
				p = client.search_post(guids=guids[:16])
				if m not in map(lambda f: f["md5"], p):
					print("Post", m, "not findable with all tags")
				guids = guids[16:]

	client = Client()
	posts = client._search_post("SPFtagguid Fimplied Fext", ["tagguid", "datatags", "implied", "ext"])
	print(len(posts), "posts")
	posts = dict(map(lambda f: (f["md5"], f), posts))
	tags = client.find_tags("EAI", "")
	print(len(tags), "tags")
	bad_images = None
	bad_thumbs = None
	stray = None
	connectivity = None
	optlow = opts.lower()
	if "i" in optlow: bad_images = check_images("I" in opts)
	if "t" in optlow: bad_thumbs = check_thumbs("T" in opts)
	if "f" in opts: stray = check_imagestore()
	if "s" in opts: connectivity = check_connectivity()

if __name__ == '__main__':
	from sys import argv
	main(argv[0], argv[1:])
